#  Created byMartin.cz
#  Copyright (c) Martin Strohalm. All rights reserved.

import numpy
import re
import json
from .. enums import *
from . matrix import Matrix
from . frame import Frame

# define constants
_CIRCLE_FORCE = (4./3.)*numpy.tan(numpy.pi/(2*4))
_SVG_COMMANDS = set('MmZzLlHhVvCcSsQqTtAa')
_SVG_COMMANDS_RE = re.compile("([MmZzLlHhVvCcSsQqTtAa])")
_SVG_COORDS_RE = re.compile("[-+]?[0-9]*\.?[0-9]+(?:[eE][-+]?[0-9]+)?")
_ANGLE_LIMIT = 0.0001


class Path(object):
    """
    Represents a complex graphical path as a sequence of standard commands. All
    of the drawing methods returns self so that they can be chained.
    """
    
    
    def __init__(self, fill_rule=EVENODD):
        """
        Initializes a new instance of Path.
        
        Args:
            fill_rule: pero.FILL_RULE
                Specifies the fill rule to be used for drawing as a value from
                pero.FILL_RULE enum. 
        """
        
        self._paths = [[]]
        self._subpath = self._paths[-1]
        self._cursor = (0, 0)
        
        self._bbox = None
        self._commands = None
        self._anchors = None
        self._handles = None
        self._start_angle = None
        self._end_angle = None
        
        self._fill_rule = fill_rule
    
    
    def __str__(self):
        """Gets standard string representation."""
        
        return "\n".join(str(x) for x in self.commands())
    
    
    @property
    def fill_rule(self):
        """
        Gets current fill rule as a value from pero.FILL_RULE enum.
        
        Returns:
            str
                Fill rule.
        """
        
        return self._fill_rule
    
    
    @fill_rule.setter
    def fill_rule(self, value):
        """
        Sets current fill rule as a value from pero.FILL_RULE enum.
        
        Args:
            value: pero.FILL_RULE
                Fill rule.
        """
        
        self._fill_rule = value
    
    
    @property
    def cursor(self):
        """
        Gets current cursor position.
        
        Returns:
            (float, float)
                Current cursor coordinates.
        """
        
        return self._cursor
    
    
    @property
    def is_multi(self):
        """Returns True if the path has multiple disconnected paths."""
        
        return len(self._paths) > 1
    
    
    @property
    def is_closed(self):
        """Returns True if the path is closed."""
        
        for path in self._paths:
            for cmd in path:
                if cmd[0] == PATH_CLOSE:
                    return True
        
        return False
    
    
    def commands(self):
        """
        Gets a sequence of all commands from all sub-paths.
        
        Returns:
            ((str, float,),)
                Sequence of commands as (pero.PATH, *values)
        """
        
        # get commands
        if self._commands is None:
            self._commands = tuple(tuple(c) for p in self._paths for c in p)
        
        return self._commands
    
    
    def anchors(self):
        """
        Gets all anchors.
        
        Returns:
            ((float, float),)
                Path anchors.
        """
        
        # get anchors
        if self._anchors is None:
            
            anchors = []
            
            # parse path
            for subpath in self._paths:
                
                # parse subpath
                origin = None
                for command in subpath:
                    
                    # get data
                    key = command[0]
                    values = command[1:]
                    
                    # close
                    if key == PATH_CLOSE:
                        continue
                    
                    # move to
                    if key == PATH_MOVE:
                        origin = values
                        continue
                    
                    # add origin
                    if origin:
                        anchors.append(origin)
                        origin = None
                    
                    # line to
                    if key == PATH_LINE:
                        anchors.append(values)
                    
                    # curve to
                    elif key == PATH_CURVE:
                        anchors.append(values[4:6])
            
            self._anchors = tuple(anchors)
        
        return self._anchors
    
    
    def handles(self):
        """
        Gets all control handles.
        
        Returns:
            ((float, float, float, float),)
                Path handles as (x, y, cx, cy).
        """
        
        # get handles
        if self._handles is None:
            
            handles = []
            
            # parse path
            for subpath in self._paths:
                
                # parse subpath
                cursor = None
                for command in subpath:
                    
                    # get data
                    key = command[0]
                    values = command[1:]
                    
                    # move to
                    if key == PATH_MOVE:
                        cursor = values
                    
                    # line to
                    elif key == PATH_LINE:
                        cursor = values
                    
                    # curve to
                    elif key == PATH_CURVE:
                        handles.append(cursor+values[0:2])
                        handles.append(values[4:6]+values[2:4])
                        cursor = values[4:6]
            
            self._handles = tuple(handles)
        
        return self._handles
    
    
    def start(self):
        """
        Gets the start point of the first sub-path.
        
        Returns:
            (float, float)
                Path start coordinates.
        """
        
        # return cursor if not path
        if not self._paths:
            return self.cursor
        
        # get first sub-path
        path = self._paths[0]
        if not path:
            return self.cursor
        
        # get start
        for command in path:
            
            # get data
            key = command[0]
            values = command[1:]
            
            # move to
            if key == PATH_MOVE:
                return values
        
        return self.cursor
    
    
    def end(self):
        """
        Gets the end point of the last sub-path.
        
        Returns:
            (float, float)
                Path end coordinates.
        """
        
        # return cursor if not path
        if not self._paths:
            return self.cursor
        
        # get first sub-path
        path = self._paths[-1]
        if not path:
            return self.cursor
        
        # get end
        for command in reversed(path):
            
            # get data
            key = command[0]
            values = command[1:]
            
            # line to
            if key == PATH_LINE:
                return values
            
            # curve to
            elif key == PATH_CURVE:
                return values[4:6]
        
        return self.cursor
    
    
    def center(self):
        """
        Gets current bounding box center.
        
        Returns:
            (float, float)
                Path center coordinates.
        """
        
        # return cursor if not path
        if not self._paths:
            return self.cursor
        
        # get center
        return self.bbox().center
    
    
    def start_angle(self):
        """
        Gets the angle of the path at the start point of the first sub-path
        in radians.
        
        Returns:
            float or None
                Start angle in radians.
        """
        
        # get angle
        if self._start_angle is None:
            
            # get first sub-path
            path = self._paths[0]
            if not path:
                return None
            
            # init points
            origin = None
            control = None
            
            # get angle
            for command in path:
                
                # get data
                key = command[0]
                values = command[1:]
                
                # move to
                if key == PATH_MOVE:
                    origin = values
                
                # line to
                elif key == PATH_LINE:
                    control = values
                
                # curve to
                elif key == PATH_CURVE:
                    control = values[:2]
                
                # calc angle
                if origin and control:
                    self._start_angle = numpy.arctan2(control[1] - origin[1], control[0] - origin[0])
                    break
        
        return self._start_angle
    
    
    def end_angle(self):
        """
        Gets the angle of the path at the end point of the last sub-path
        in radians.
        
        Returns:
            float or None
                End angle in radians.
        """
        
        # get angle
        if self._end_angle is None:
            
            # get last sub-path
            path = self._paths[-1]
            if not path:
                return None
            
            # init points
            origin = None
            control = None
            
            # get angle
            for command in reversed(path):
                
                # get data
                key = command[0]
                values = command[1:]
                
                # move to
                if key == PATH_MOVE:
                    origin = values
                
                # line to
                elif key == PATH_LINE:
                    if control:
                        origin = values
                    else:
                        control = values
                
                # curve to
                elif key == PATH_CURVE:
                    if control:
                        origin = values[4:6]
                    else:
                        origin = values[2:4]
                        control = values[4:6]
                
                # calc angle
                if origin and control:
                    self._end_angle = numpy.arctan2(control[1] - origin[1], control[0] - origin[0])
                    break
        
        return self._end_angle
    
    
    def bbox(self):
        """
        Calculates current path bounding box.
        
        Returns:
            pero.Frame or None
                Path bounding box or None if path doesn't contain any command.
        """
        
        # calc box
        if self._bbox is None:
            
            # init origin
            x1 = 0
            y1 = 0
            
            # add commands
            for command in self.commands():
                
                # get data
                key = command[0]
                values = command[1:]
                
                # close
                if key == PATH_CLOSE:
                    continue
                
                # move to
                if key == PATH_MOVE:
                    x1, y1 = values
                    continue
                
                # init frame
                if self._bbox is None:
                    self._bbox = Frame(x1, y1)
                
                # line to
                if key == PATH_LINE:
                    self._bbox.extend(x1, y1)
                    x1, y1 = values
                    self._bbox.extend(x1, y1)
                
                # curve to
                elif key == PATH_CURVE:
                    cx1, cy1, cx2, cy2, x2, y2 = values
                    bbox = Path.curve_bbox(x1, y1, cx1, cy1, cx2, cy2, x2, y2)
                    self._bbox.extend(*bbox.rect)
                    x1, y1 = x2, y2
        
        return self._bbox.clone()
    
    
    def json(self):
        """
        Gets current path as JSON dump.
        
        Returns:
            str
                JSON dump.
        """
        
        return json.dumps({
            "fill_rule": self.fill_rule,
            "commands": self.commands()})
    
    
    def svg(self, indent="", rounding=None):
        """
        Gets current path as SVG commands.
        
        Args:
            indent: str
                XML indentation.
            
            rounding: int or None
                Rounding applied to all coordinates.
        
        Returns:
            str
                SVG commands
        """
        
        full_svg = []
        for path in self._paths:
            
            path_svg = []
            for commands in path:
                key = commands[0]
                
                values = commands[1:]
                if rounding is not None:
                    values = (round(x, rounding) for x in values)
                
                command = " ".join(str(x) for x in values) if values else ""
                path_svg.append(key+command)
            
            full_svg.append(" ".join(path_svg))
        
        indent = "\n" + indent
        
        return indent + indent.join(full_svg)
    
    
    def dirty(self):
        """Resets internal cached values."""
        
        self._bbox = None
        self._commands = None
        self._anchors = None
        self._handles = None
        self._start_angle = None
        self._end_angle = None
    
    
    def close(self):
        """
        Adds a line segment from current point to the beginning of the current
        sub-path. A new sub-path will be initialized.
        
        Returns:
            pero.Path
                Returns self so that the commands can be chained.
        """
        
        # check if sub-path exists
        if not self._subpath:
            return self
        
        # close current sub-path
        self._subpath.append([PATH_CLOSE])
        
        # move cursor
        self._cursor = self._subpath[0][1:3]
        
        # init new sub-path
        self._subpath = []
        self._paths.append(self._subpath)
        
        # make dirty
        self.dirty()
        
        return self
    
    
    def move_to(self, x, y, relative=False):
        """
        Begins a new sub-path by moving current point to given coordinates.
        
        Args:
            x: int or float
                X-coordinate of the point to move to.
            
            y: int or float
                Y-coordinate of the point to move to.
            
            relative: bool
                If set to True given coordinates are considered as relative to
                current point.
        
        Returns:
            pero.Path
                Returns self so that the commands can be chained.
        """
        
        # get absolute coordinates
        if relative:
            x += self._cursor[0]
            y += self._cursor[1]
        
        # move cursor
        self._cursor = (x, y)
        
        # init new sub-path
        if self._subpath:
            self._subpath = []
            self._paths.append(self._subpath)
        
        return self
    
    
    def line_to(self, x=None, y=None, relative=False):
        """
        Adds a straight line from the current point to given coordinates.
        
        Args:
            x: int or float
                X-coordinate of the end point.
            
            y: int or float
                Y-coordinate of the end point.
            
            relative: bool
                If set to True given coordinates are considered as relative to
                current point.
        
        Returns:
            pero.Path
                Returns self so that the commands can be chained.
        """
        
        # check coords
        if x is None and y is None:
            return
        
        # init sub-path if necessary
        if not self._subpath:
            self._subpath.append([PATH_MOVE, self._cursor[0], self._cursor[1]])
        
        # get absolute coordinates
        if relative:
            if x is not None:
                x += self._cursor[0]
            if y is not None:
                y += self._cursor[1]
        
        # vertical line
        if x is None:
            x = self._cursor[0]
        
        # horizontal line
        if y is None:
            y = self._cursor[1]
        
        # add line
        self._subpath.append([PATH_LINE, x, y])
        
        # move cursor
        self._cursor = (x, y)
        
        # make dirty
        self.dirty()
        
        return self
    
    
    def ray_to(self, angle, length):
        """
        Adds a straight line from the current point by given angle and length.
        
        Args:
            angle: int or float
                Line angle in radians.
            
            length: int or float
                Line length.
        
        Returns:
            pero.Path
                Returns self so that the commands can be chained.
        """
        
        # get previous end point
        x1, y1 = self._cursor
        
        # calc end point
        x2 = x1 + length * numpy.cos(angle)
        y2 = y1 + length * numpy.sin(angle)
        
        # add line
        self.line_to(x2, y2)
        
        return self
    
    
    def curve_to(self, cx1, cy1, cx2, cy2, x, y, relative=False):
        """
        Adds a cubic Bezier curve from the current point using control points
        and end point.
        
        Args:
            cx1: int or float
                X-coordinate of the start control point.
            
            cy1: int or float
                Y-coordinate of the start control point.
            
            cx2: int or float
                X-coordinate of the end control point.
            
            cy2: int or float
                Y-coordinate of the end control point.
            
            x: int or float
                X-coordinate of the end point.
            
            y: int or float
                Y-coordinate of the end point.
            
            relative: bool
                If set to True given coordinates are considered as relative to
                current point.
        
        Returns:
            pero.Path
                Returns self so that the commands can be chained.
        """
        
        # init sub-path if necessary
        if not self._subpath:
            self._subpath.append([PATH_MOVE, self._cursor[0], self._cursor[1]])
        
        # get absolute coordinates
        if relative:
            cx1 += self._cursor[0]
            cy1 += self._cursor[1]
            cx2 += self._cursor[0]
            cy2 += self._cursor[1]
            x += self._cursor[0]
            y += self._cursor[1]
        
        # add curve
        self._subpath.append([PATH_CURVE, cx1, cy1, cx2, cy2, x, y])
        
        # move cursor
        self._cursor = (x, y)
        
        # make dirty
        self.dirty()
        
        return self
    
    
    def curve_s_to(self, cx2, cy2, x, y, relative=False):
        """
        Adds a smooth cubic Bezier curve from the current point, using
        reflection of the previous control point and given end point and control
        point.
        
        Args:
            cx2: int or float
                X-coordinate of the end control point.
            
            cy2: int or float
                Y-coordinate of the end control point.
            
            x: int or float
                X-coordinate of the end point.
            
            y: int or float
                Y-coordinate of the end point.
            
            relative: bool
                If set to True given coordinates are considered as relative to
                current point.
        
        Returns:
            pero.Path
                Returns self so that the commands can be chained.
        """
        
        # get absolute coordinates
        if relative:
            cx2 += self._cursor[0]
            cy2 += self._cursor[1]
            x += self._cursor[0]
            y += self._cursor[1]
        
        # get previous control point
        cx1, cy1 = self._cursor
        if self._subpath and self._subpath[-1][0] == PATH_CURVE:
            cx1, cy1 = self._subpath[-1][3:5]
            cx1 += 2*(self._cursor[0] - cx1)
            cy1 += 2*(self._cursor[1] - cy1)
        
        # add curve
        self.curve_to(cx1, cy1, cx2, cy2, x, y)
        
        return self
    
    
    def quad_to(self, cx, cy, x, y, relative=False):
        """
        Adds a quadratic Bezier curve from the current point using control and
        end point.
        
        Args:
            cx: int or float
                X-coordinate of the control point.
            
            cy: int or float
                Y-coordinate of the control point.
            
            x: int or float
                X-coordinate of the end point.
            
            y: int or float
                Y-coordinate of the end point.
            
            relative: bool
                If set to True given coordinates are considered as relative to
                current point.
        
        Returns:
            pero.Path
                Returns self so that the commands can be chained.
        """
        
        # get absolute coordinates
        if relative:
            x += self._cursor[0]
            y += self._cursor[1]
            cx += self._cursor[0]
            cy += self._cursor[1]
        
        # get control points
        cx1 = cx - (cx - self._cursor[0])/3.
        cy1 = cy - (cy - self._cursor[1])/3.
        
        cx2 = cx - (cx - x)/3.
        cy2 = cy - (cy - y)/3.
        
        # add curve
        self.curve_to(cx1, cy1, cx2, cy2, x, y)
        
        return self
    
    
    def quad_s_to(self, x, y, relative=False):
        """
        Adds a smooth quadratic Bezier curve from the current point, using
        reflection of the previous control point and given end point.
        
        Args:
            x: int or float
                X-coordinate of the end point.
            
            y: int or float
                Y-coordinate of the end point.
            
            relative: bool
                If set to True given coordinates are considered as relative to
                current point.
        
        Returns:
            pero.Path
                Returns self so that the commands can be chained.
        """
        
        # get absolute coordinates
        if relative:
            x += self._cursor[0]
            y += self._cursor[1]
        
        # get previous control point
        cx1, cy1 = self._cursor
        if self._subpath and self._subpath[-1][0] == PATH_CURVE:
            cx1, cy1 = self._subpath[-1][3:5]
            cx1 += 2*(self._cursor[0] - cx1)
            cy1 += 2*(self._cursor[1] - cy1)
        
        # get previous end point
        x1, y1 = self._cursor
        
        # get quad control point
        cx = .5*(3*cx1 - x1)
        cy = .5*(3*cy1 - y1)
        
        # get end control point
        cx2 = cx - (cx-x)/3.
        cy2 = cy - (cy-y)/3.
        
        # add curve
        self.curve_to(cx1, cy1, cx2, cy2, x, y)
        
        return self
    
    
    def bow_to(self, x, y, radius, large=False, clockwise=True, relative=False):
        """
        Adds a circular arc from the current point, using given radius and end
        point coordinates. One of the four existing solutions is chosen
        according to the 'large' and 'clockwise' parameters.
        
        If the radius is too small for specified end points the theoretical
        minimum is used, which equals half of the end points distance.
        
        Args:
            x: int or float
                X-coordinate of the end point.
            
            y: int or float
                Y-coordinate of the end point.
            
            radius: int or float
                Radius of the arc.
            
            large: bool
                Specifies which of the possible arcs will be drawn according
                to its length.
            
            clockwise: bool
                Specifies which of the possible arcs will be drawn according to
                drawing direction. If set to True the clockwise arc is drawn,
                otherwise the anti-clockwise.
            
            relative: bool
                If set to True given coordinates are considered as relative to
                current point.
        
        Returns:
            pero.Path
                Returns self so that the commands can be chained.
        """
        
        # get absolute coordinates
        if relative:
            x += self._cursor[0]
            y += self._cursor[1]
        
        # get start point
        x1, y1 = self._cursor
        
        # get points distance and angle
        dist = 0.5 * numpy.sqrt((x - x1) ** 2 + (y - y1) ** 2)
        angle = numpy.arctan2(y - y1, x - x1)
        
        # check min radius
        radius = max(radius, dist)
        
        # calc sin/cos
        sin = numpy.sin(angle)
        cos = numpy.cos(angle)
        
        # get origins
        c = numpy.sqrt(radius ** 2 - dist ** 2)
        c1x = x1 + dist * cos - c * sin
        c1y = y1 + dist * sin + c * cos
        c2x = x1 + dist * cos + c * sin
        c2y = y1 + dist * sin - c * cos
        
        # set circles
        if angle >= 0:
            small = c1x, c1y
            big = c2x, c2y
        else:
            big = c2x, c2y
            small = c1x, c1y
        
        # apply direction
        if not clockwise:
            small, big = big, small
        
        # select final circle
        cx, cy = big if large else small
        
        # get end angle
        end_angle = numpy.arctan2(y - cy, x - cx)
        
        # add arc
        self.arc_around(cx, cy, end_angle, clockwise=clockwise)
        
        return self
    
    
    def arc_to(self, cx, cy, x, y, radius=None, relative=False, finalize=False, limit=True):
        """
        Adds a circular arc from the current point, using given radius and
        control and end point coordinates. This is achieved by adding additional
        points (as needed) between the endpoints to reach smoothly the desired
        arc radius.
        
        The initial strait segment between start point and the first new point
        is added automatically. By default the final strait segment is omitted
        and the path ends at the last newly added point. This behavior can be
        changed by setting the 'finalize' to True.
        
        If the radius is too big for given points the final curve creates
        characteristic flipping effect with sharp corners. This behavior can be
        avoided by setting the 'limit' to True to limit the radius to maximum
        value to get smooth curve. The same value is also used automatically
        if radius is not specified.
        
        Args:
            cx: int or float
                X-coordinate of the control point.
            
            cy: int or float
                Y-coordinate of the control point.
            
            x: int or float
                X-coordinate of the end point.
            
            y: int or float
                Y-coordinate of the end point.
            
            radius: int or float
                Radius of the arc.
            
            relative: bool
                If set to True given coordinates are considered as relative to
                current point.
            
            finalize: bool
                If set to True the final strait segment to the end point is
                added automatically.
            
            limit: bool
                If set to True the maximum allowed radius is used to avoid the
                curve flipping effect.
        
        Returns:
            pero.Path
                Returns self so that the commands can be chained.
        """
        
        # get absolute coordinates
        if relative:
            x += self._cursor[0]
            y += self._cursor[1]
            cx += self._cursor[0]
            cy += self._cursor[1]
        
        # get start point
        x1, y1 = self._cursor
        
        # check control point
        if (x1 == cx and y1 == cy) or (x == cx and y == cy) or radius == 0:
            if finalize: self.line_to(x, y)
            else: self.line_to(cx, cy)
            return self
        
        # check same point
        if x1 == x and y1 == y:
            if finalize: self.line_to(x, y)
            return self
        
        # check points in line
        if abs((cy - y1) * (x - x1) - (cx - x1) * (y - y1)) < 1e-5:
            if finalize: self.line_to(x, y)
            else: self.line_to(cx, cy)
            return self
        
        # get total angle
        angle = numpy.arctan2(cy-y1, cx-x1) - numpy.arctan2(cy-y, cx-x)
        
        # get distances
        x1c = numpy.sqrt((cx - x1)**2 + (cy - y1)**2)
        x2c = numpy.sqrt((cx - x)**2 + (cy - y)**2)
        
        # get limits
        min_dist = min(x1c, x2c)
        max_radius = min_dist * abs(numpy.tan(0.5*angle))
        
        # use max radius if not set
        if radius is not None:
            dist = radius / abs(numpy.tan(0.5*angle))
        else:
            dist = min_dist
            radius = max_radius
        
        # check limits
        if limit and (dist > x1c or dist > x2c):
            dist = min_dist
            radius = max_radius
        
        # get arc endpoints
        ax1 = cx - (cx - x1) * dist / x1c
        ay1 = cy - (cy - y1) * dist / x1c
        ax2 = cx - (cx - x) * dist / x2c
        ay2 = cy - (cy - y) * dist / x2c
        
        # get arc origin
        dx = cx * 2 - ax1 - ax2
        dy = cy * 2 - ay1 - ay2
        ac = numpy.sqrt(dx**2 + dy**2)
        oc = numpy.sqrt(radius**2 + dist**2)
        
        ox = cx - dx * oc / ac
        oy = cy - dy * oc / ac
        
        # get end angle
        end_angle = numpy.arctan2(ay2-oy, ax2-ox)
        
        # get arc direction
        clockwise = True
        if -numpy.pi < angle < 0 or numpy.pi < angle:
            clockwise = False
        
        # add initial line
        if x1 != ax1 or y1 != ay1:
            self.line_to(ax1, ay1)
        
        # add arc
        self.arc_around(ox, oy, end_angle, clockwise=clockwise)
        
        # add final line
        if finalize and (x != ax2 or y != ay2):
            self.line_to(x, y)
        
        return self
    
    
    def arc_around(self, x, y, end_angle, clockwise=True, relative=False):
        """
        Adds a circular arc from the current point around given center
        coordinates up to specified end angle.
        
        Args:
            x: int or float
                X-coordinate of the center.
            
            y: int or float
                Y-coordinate of the center.
            
            end_angle: float
                End angle in radians.
            
            clockwise: bool
                Specifies the direction of drawing. If set to True, the arc
                will be drawn in the clockwise direction.
            
            relative: bool
                If set to True given coordinates are considered as relative to
                current point.
        
        Returns:
            pero.Path
                Returns self so that the commands can be chained.
        """
        
        # get absolute coordinates
        if relative:
            x += self._cursor[0]
            y += self._cursor[1]
        
        # init constants
        pi2 = numpy.pi * 2
        pi05 = numpy.pi * 0.5
        
        # get start point
        x1, y1 = self._cursor
        
        # get radius
        radius = numpy.sqrt((x - x1) ** 2 + (y - y1) ** 2)
        
        # get angles
        start_angle = numpy.arctan2(y1 - y, x1 - x) % pi2
        end_angle = end_angle % pi2
        total_angle = numpy.abs(end_angle - start_angle) % pi2
        
        if start_angle > end_angle:
            total_angle = pi2 - total_angle
        
        # change direction
        direction = 1
        if not clockwise:
            total_angle = pi2 - total_angle
            direction = -1
        
        # init start
        a1 = start_angle
        
        # create segments
        while total_angle > _ANGLE_LIMIT:
            # calc segment angle
            a2 = a1 + direction * min(total_angle, pi05)
            angle = (a2 - a1)
            
            # calc force
            force = radius * (4. / 3.) * numpy.tan(numpy.pi / (2 * pi2 / angle))
            
            # calc coordinates
            x2 = x + radius * numpy.cos(a2)
            y2 = y + radius * numpy.sin(a2)
            cx1 = x1 + force * numpy.cos(a1 + pi05)
            cy1 = y1 + force * numpy.sin(a1 + pi05)
            cx2 = x2 + force * numpy.cos(a2 - pi05)
            cy2 = y2 + force * numpy.sin(a2 - pi05)
            
            # add curve
            self.curve_to(cx1, cy1, cx2, cy2, x2, y2)
            
            # update values
            total_angle -= numpy.abs(angle)
            a1 = a2
            x1 = x2
            y1 = y2
        
        return self
    
    
    def arc(self, x, y, radius, start_angle, end_angle, clockwise=True):
        """
        Adds an arc as a new sub-path of given radius centered around given
        coordinates.
        
        Args:
            x: int or float
                X-coordinate of the center.
            
            y: int or float
                Y-coordinate of the center.
            
            radius: int or float
                Radius of the arc.
            
            start_angle: float
                Start angle in radians.
            
            end_angle: float
                End angle in radians.
            
            clockwise: bool
                Specifies the direction of drawing. If set to True, the arc
                will be drawn in the clockwise direction.
        
        Returns:
            pero.Path
                Returns self so that the commands can be chained.
        """
        
        # move to start
        x1 = x + radius * numpy.cos(start_angle)
        y1 = y + radius * numpy.sin(start_angle)
        self.move_to(x1, y1)
        
        # add arc
        self.arc_around(x, y, end_angle, clockwise)
        
        return self
    
    
    def circle(self, x, y, radius, relative=False):
        """
        Adds a circle as a new closed sub-path of given radius centered around
        given coordinates. Current cursor point stays at the circle center.
        
        Args:
            x: int or float
                X-coordinate of the center.
            
            y: int or float
                Y-coordinate of the center.
            
            radius: int or float
                Radius of the circle.
            
            relative: bool
                If set to True center coordinates are considered as relative to
                current point.
        
        Returns:
            pero.Path
                Returns self so that the commands can be chained.
        """
        
        # get absolute coordinates
        if relative:
            x += self._cursor[0]
            y += self._cursor[1]
        
        # calc control handle length
        force = radius * _CIRCLE_FORCE
        
        # create circle
        self.move_to(x, y-radius)
        self.curve_to(x+force, y-radius, x+radius, y-force, x+radius, y)
        self.curve_to(x+radius, y+force, x+force, y+radius, x, y+radius)
        self.curve_to(x-force, y+radius, x-radius, y+force, x-radius, y)
        self.curve_to(x-radius, y-force, x-force, y-radius, x, y-radius)
        self.close()
        
        # move cursor to center
        self._cursor = (x, y)
        
        return self
    
    
    def ellipse(self, x, y, width, height, relative=False):
        """
        Adds an ellipse as a new closed sub-path of given size centered around
        given coordinates. Current cursor point stays at the ellipse center.
        
        Args:
            x: int or float
                X-coordinate of the center.
            
            y: int or float
                Y-coordinate of the center.
            
            width: int or float
                Full width of the ellipse.
            
            height: int or float
                Full height of the ellipse.
            
            relative: bool
                If set to True center coordinates are considered as relative to
                current point.
        """
        
        # get absolute coordinates
        if relative:
            x += self._cursor[0]
            y += self._cursor[1]
        
        # create ellipse
        path = Path().circle(0, 0, 0.5*height)
        scale = float(width) / height
        matrix = Matrix().scale(scale, 1).translate(x, y)
        path.transform(matrix)
        
        # add to current path
        self.path(path)
        
        # move cursor to center
        self._cursor = (x, y)
        
        return self
    
    
    def rect(self, x, y, width, height, radius=0, relative=False):
        """
        Adds a rectangle as a new closed sub-path of given size and the top-left
        origin coordinates. Current cursor point stays at the top left corner.
        
        Args:
            x: int or float
                X-coordinate of the top left corner.
            
            y: int or float
                Y-coordinate of the top left corner.
            
            width: int or float
                Full rectangle width.
            
            height: int or float
                Full rectangle height.
            
            radius: int or float or collection of four int/float
                Radius of the corners as a single value or separate value for
                each corner individually, starting top-left.
            
            relative: bool
                If set to True origin coordinates are considered as relative to
                current point.
        
        Returns:
            pero.Path
                Returns self so that the commands can be chained.
        """
        
        # get absolute coordinates
        if relative:
            x += self._cursor[0]
            y += self._cursor[1]
        
        # create standard rectangle
        if not radius:
            
            self.move_to(x, y)
            self.line_to(width, 0, relative=True)
            self.line_to(0, height, relative=True)
            self.line_to(-width, 0, relative=True)
            self.close()
            
            return self
        
        # use separate radii
        if isinstance(radius, (list, tuple)):
            r1, r2, r3, r4 = radius
        else:
            r1, r2, r3, r4 = radius, radius, radius, radius
        
        # calc control handle length
        f1 = r1 * _CIRCLE_FORCE
        f2 = r2 * _CIRCLE_FORCE
        f3 = r3 * _CIRCLE_FORCE
        f4 = r4 * _CIRCLE_FORCE
        
        self.move_to(x+width-r2, y)
        self.curve_to(f2, 0, r2, r2-f2, r2, r2, relative=True)
        self.line_to(0, height-r2-r3, relative=True)
        self.curve_to(0, f3, -r3+f3, r3, -r3, r3, relative=True)
        self.line_to(-width+r3+r4, 0, relative=True)
        self.curve_to(-f4, 0, -r4, -r4+f4, -r4, -r4, relative=True)
        self.line_to(0, -height+r4+r1, relative=True)
        self.curve_to(0, -f1, r1-f1, -r1, r1, -r1, relative=True)
        self.close()
        
        # move cursor to origin
        self._cursor = (x, y)
        
        return self
    
    
    def polygon(self, points, relative=False):
        """
        Adds a polygon as a new closed sub-path defined by given points.
        
        Args:
            points: list of (float, float)
                Collection of x,y coordinates of the points.
            
            relative: bool
                If set to True all coordinates are considered as relative to
                previous point.
        
        Returns:
            pero.Path
                Returns self so that the commands can be chained.
        """
        
        # move to start
        x, y = points[0]
        self.move_to(x, y, relative=relative)
        
        # add points
        for point in points[1:]:
            self.line_to(point[0], point[1], relative=relative)
        
        # close polygon
        self.close()
        
        return self
    
    
    def path(self, path):
        """
        Adds a given path as a new closed sub-path.
        
        Args:
            path: pero.Path
                Path to be added.
        
        Returns:
            pero.Path
                Returns self so that the commands can be chained.
        """
        
        # add commands
        for command in path.commands():
            
            # get data
            key = command[0]
            values = command[1:]
            
            # close
            if key == PATH_CLOSE:
                self.close()
            
            # move to
            elif key == PATH_MOVE:
                self.move_to(*values)
            
            # line to
            elif key == PATH_LINE:
                self.line_to(*values)
            
            # curve to
            elif key == PATH_CURVE:
                self.curve_to(*values)
        
        return self
    
    
    def transform(self, matrix):
        """
        Transforms each command by transformation matrix.
        
        Args:
            matrix: pero.Matrix
                Transformation matrix to apply.
        
        Returns:
            pero.Path
                Returns self so that the commands can be chained.
        """
        
        # transform commands
        for subpath in self._paths:
            for command in subpath:
                
                # get data
                key = command[0]
                values = command[1:]
                
                # move to
                if key == PATH_MOVE:
                    command[1:3] = matrix.transform(values[0], values[1])
                
                # line to
                elif key == PATH_LINE:
                    command[1:3] = matrix.transform(values[0], values[1])
                
                # curve to
                elif key == PATH_CURVE:
                    command[1:3] = matrix.transform(values[0], values[1])
                    command[3:5] = matrix.transform(values[2], values[3])
                    command[5:7] = matrix.transform(values[4], values[5])
        
        # transform cursor
        self._cursor = tuple(matrix.transform(self._cursor[0], self._cursor[1]))
        
        # make dirty
        self.dirty()
        
        return self
    
    
    def transformed(self, matrix, fill_rule=None):
        """
        Creates a clone of current path and applies the transformation.
        
        Args:
            matrix: pero.Matrix
                Transformation matrix to apply.
            
            fill_rule: pero.FILL_RULE
                Specifies the fill rule to be used for drawing as a value from
                pero.FILL_RULE enum. If set to None, current path value is used.
        
        Returns:
            pero.Path
                Transformed path as a new instance.
        """
        
        return self.clone(fill_rule).transform(matrix)
    
    
    def split(self):
        """
        Splits current path into individual sub-paths.
        
        Returns:
            (pero.Path,)
                Individual paths.
        """
        
        # just one path
        if len(self._paths) == 1:
            (self.clone(),)
        
        # split paths
        return tuple(Path.from_commands(x) for x in self._paths if x)
    
    
    def symbol(self, fill_rule=None):
        """
        Creates a clone of current path centered at 0,0 and rescaled to fit into
        1 x 1 square.
        
        Args:
            fill_rule: pero.FILL_RULE
                Specifies the fill rule to be used for drawing as a value from
                pero.FILL_RULE enum. If set to None, current path value is used.
        
        Returns:
            pero.Path
                Symbol path.
        """
        
        # get current bounding box
        bbox = self.bbox()
        
        # init matrix
        matrix = Matrix()
        
        # center around zero
        matrix.translate(-bbox.cx, -bbox.cy)
        
        # scale to 1 x 1
        scale = 1. / max(bbox.width, bbox.height)
        matrix.scale(scale, scale)
        
        # apply to path
        return self.clone(fill_rule).transform(matrix)
    
    
    def clone(self, fill_rule=None):
        """
        Creates a clone of current path.
        
        Args:
            fill_rule: pero.FILL_RULE
                Specifies the fill rule to be used for drawing as a value from
                pero.FILL_RULE enum. If set to None, current path value is used.
        
        Returns:
            pero.Path
                Direct clone of current path.
        """
        
        # get fill rule
        if fill_rule is None:
            fill_rule = self._fill_rule
        
        # clone commands
        path = Path.from_commands(self.commands(), fill_rule)
        
        # set cursor
        path._cursor = self._cursor
        
        return path
    
    
    @staticmethod
    def curve_bbox(x1, y1, cx1, cy1, cx2, cy2, x2, y2):
        """
        Gets bounding box of Bezier curve.
        
        Args:
            x1: int or float
                X-coordinate of the start point.
            
            y1: int or float
                Y-coordinate of the start point.
            
            cx1: int or float
                X-coordinate of the start control point.
            
            cy1: int or float
                Y-coordinate of the start control point.
            
            cx2: int or float
                X-coordinate of the end control point.
            
            cy2: int or float
                Y-coordinate of the end control point.
            
            x2: int or float
                X-coordinate of the end point.
            
            y2: int or float
                Y-coordinate of the end point.
        
        Returns:
            tuple of int or float
                Curve bounding box as top-left x, y, width and height.
        """
        
        # init box
        bbox = Frame(x1, y1)
        bbox.extend(x2, y2)
        
        # init points
        p1 = (x1, y1)
        c1 = (cx1, cy1)
        c2 = (cx2, cy2)
        p2 = (x2, y2)
        
        for i in range(2):
            
            def f(t):
                return numpy.power(1 - t, 3) * p1[i] \
                    + 3 * numpy.power(1 - t, 2) * t * c1[i] \
                    + 3 * (1 - t) * numpy.power(t, 2) * c2[i] \
                    + numpy.power(t, 3) * p2[i]
            
            b = 6 * p1[i] - 12 * c1[i] + 6 * c2[i]
            a = -3 * p1[i] + 9 * c1[i] - 9 * c2[i] + 3 * p2[i]
            c = 3 * c1[i] - 3 * p1[i]
            
            if a == 0:
                
                if b == 0:
                    continue
                
                t = -c / b
                if 0 < t < 1:
                    if i == 0:
                        bbox.extend(x=f(t))
                    elif i == 1:
                        bbox.extend(y=f(t))
                
                continue
            
            b2ac = numpy.power(b, 2) - 4 * c * a
            if b2ac < 0:
                continue
            
            t1 = (-b + numpy.sqrt(b2ac)) / (2 * a)
            if 0 < t1 < 1:
                if i == 0:
                    bbox.extend(x=f(t1))
                elif i == 1:
                    bbox.extend(y=f(t1))
            
            t2 = (-b - numpy.sqrt(b2ac)) / (2 * a)
            if 0 < t2 < 1:
                if i == 0:
                    bbox.extend(x=f(t2))
                elif i == 1:
                    bbox.extend(y=f(t2))
            
        return bbox
    
    
    @staticmethod
    def from_commands(commands, fill_rule=EVENODD):
        """
        Creates a new path from given commands.
        
        Args:
            commands: ((pero.PATH, float,),)
                Sequence of commands as (key, *values), where key must be a
                value from the pero.PATH enum.
            
            fill_rule: pero.FILL_RULE
                Specifies the fill rule to be used for drawing as a value from
                pero.FILL_RULE enum.
        
        Returns:
            pero.Path
                Path created from given commands.
        """
        
        # init path
        path = Path(fill_rule)
        
        # add commands
        for command in commands:
            
            # get data
            key = command[0]
            values = command[1:]
            
            # close
            if key == PATH_CLOSE:
                path.close()
            
            # move to
            elif key == PATH_MOVE:
                path.move_to(*values)
            
            # line to
            elif key == PATH_LINE:
                path.line_to(*values)
            
            # curve to
            elif key == PATH_CURVE:
                path.curve_to(*values)
        
        return path
    
    
    @staticmethod
    def from_json(dump):
        """
        Creates a new path from given JSON dump.
        
        Args:
            dump: str
                JSON string representing the path.
        
        Returns:
            pero.Path
                Path created from given JSON dump.
        """
        
        # load json
        if not isinstance(dump, dict):
            dump = json.loads(dump)
        
        # init path
        path = Path(fill_rule=dump.get("fill_rule", EVENODD))
        
        # add commands
        for command in dump.get("commands", []):
            
            # get data
            key = command[0]
            values = command[1:]
            
            # close
            if key == PATH_CLOSE:
                path.close()
            
            # move to
            elif key == PATH_MOVE:
                path.move_to(*values)
            
            # line to
            elif key == PATH_LINE:
                path.line_to(*values)
            
            # curve to
            elif key == PATH_CURVE:
                path.curve_to(*values)
        
        return path
    
    
    @staticmethod
    def from_svg(svg, fill_rule=EVENODD):
        """
        Creates a new path by parsing given SVG path definition.
        
        Args:
            svg: str
                SVG commands.
            
            fill_rule: pero.FILL_RULE
                Specifies the fill rule to be used for drawing as a value from
                pero.FILL_RULE enum.
        
        Returns:
            pero.Path
                Path created from given SVG commands.
        """
        
        # init path
        path = Path(fill_rule)
        
        # parse commands
        command = None
        for x in _SVG_COMMANDS_RE.split(svg):
            
            # close
            if x == 'Z' or x == 'z':
                path.close()
                continue
            
            # get command
            if x in _SVG_COMMANDS:
                command = x
                continue
            
            # get coords
            coords = tuple(map(float, _SVG_COORDS_RE.findall(x)))
            if not coords:
                continue
            
            # move to
            if command == 'M':
                path.move_to(*coords)
            elif command == 'm':
                path.move_to(*coords, relative=True)
            
            # line to
            elif command == 'L':
                path.line_to(*coords)
            elif command == 'l':
                path.line_to(*coords, relative=True)
            
            # horizontal line to
            elif command == 'H':
                path.line_to(x=coords[0])
            elif command == 'h':
                path.line_to(x=coords[0], relative=True)
            
            # vertical line to
            elif command == 'V':
                path.line_to(y=coords[0])
            elif command == 'v':
                path.line_to(y=coords[0], relative=True)
            
            # curve to
            elif command == 'C':
                path.curve_to(*coords)
            elif command == 'c':
                path.curve_to(*coords, relative=True)
            
            # smooth curve to
            elif command == 'S':
                path.curve_s_to(*coords)
            elif command == 's':
                path.curve_s_to(*coords, relative=True)
            
            # quad to
            elif command == 'Q':
                path.quad_to(*coords)
            elif command == 'q':
                path.quad_to(*coords, relative=True)
            
            # smooth quad to
            elif command == 'T':
                path.quad_s_to(*coords)
            elif command == 't':
                path.quad_s_to(*coords, relative=True)
            
            # unknown
            else:
                raise ValueError("Unsupported path command! -> %s" % command)
        
        return path
    
    
    @staticmethod
    def from_bezier(curve, fill_rule=EVENODD):
        """
        Creates a new path from given Bezier curve.
        
        Args:
            curve: pero.Bezier
                Bezier curve.
            
            fill_rule: pero.FILL_RULE
                Specifies the fill rule to be used for drawing as a value from
                pero.FILL_RULE enum.
        
        Returns:
            pero.Path
                Path created from given Bezier.
        """
        
        # get curve coords
        coords = curve.coords
        
        # make path
        path = Path(fill_rule)
        path.move_to(coords[0], coords[1])
        path.curve_to(*coords[2:])
        
        return path
    
    
    @staticmethod
    def make_star(rays, x=0, y=0, outer_radius=.5, inner_radius=.25, angle=0, fill_rule=EVENODD):
        """
        Creates a closed star-like path.
        
        Args:
            rays: int
                Number of rays.
            
            x: int or float
                X-coordinate of the center.
            
            y: int or float
                Y-coordinate of the center.
            
            outer_radius: int or float
                Outer radius of the rays.
            
            inner_radius: int or float
                Inner radius of the rays.
            
            angle: float
                Angle in radians.
            
            fill_rule: pero.FILL_RULE
                Specifies the fill rule to be used for drawing as a value from
                pero.FILL_RULE enum.
        
        Returns:
            pero.Path
                A new star-like path.
        """
        
        # calc vertices
        sides = rays * 2
        theta = (2*numpy.pi/sides * numpy.arange(sides + 1)) - numpy.pi / 2.0
        
        r = numpy.full(sides + 1, outer_radius, dtype=float)
        r[1::2] = inner_radius
        
        vertices = numpy.stack((r*numpy.cos(theta), r*numpy.sin(theta)), axis=1)
        vertices += numpy.array((x, y))
        
        # make path
        path = Path(fill_rule).polygon(vertices[:-1])
        
        # apply angle
        if angle:
            matrix = Matrix().rotate(angle, x=x, y=y)
            path.transform(matrix)
        
        return path
    
    
    @staticmethod
    def make_ngon(sides, x=0, y=0, radius=.5, angle=0, fill_rule=EVENODD):
        """
        Creates a closed regular polygon path.
        
        Args:
            sides: int
                Number of sides/vertices.
            
            x: int or float
                X-coordinate of the center.
            
            y: int or float
                Y-coordinate of the center.
            
            radius: int or float
                Radius of the polygon.
            
            angle: float
                Angle in radians.
            
            fill_rule: pero.FILL_RULE
                Specifies the fill rule to be used for drawing as a value from
                pero.FILL_RULE enum.
        
        Returns:
            pero.Path
                A new regular polygon path.
        """
        
        # calc vertices
        theta = (2*numpy.pi/sides * numpy.arange(sides + 1)) - numpy.pi / 2.0
        r = numpy.full(sides + 1, radius, dtype=float)
        
        vertices = numpy.stack((r*numpy.cos(theta), r*numpy.sin(theta)), axis=1)
        vertices += numpy.array((x, y))
        
        # make path
        path = Path(fill_rule).polygon(vertices[:-1])
        
        # apply angle
        if angle:
            matrix = Matrix().rotate(angle, x=x, y=y)
            path.transform(matrix)
        
        return path



#  Created byMartin.cz
#  Copyright (c) Martin Strohalm. All rights reserved.

from .. enums import *
from .. properties import *
from .. drawing import Frame, Path
from . glyph import Glyph
from . markers import Marker


class Legend(Glyph):
    """
    Abstract base class for various types of legend items glyphs.
    
    The pero.Legend classes can be used directly to draw legend items or as
    descriptor to create a pero.Legend instances from real data and using the
    'clone' method and a data source.
    
    Properties:
        
        bull_x: int, float or callable
            Specifies the x-coordinate of the bullet center.
        
        bull_y: int, float or callable
            Specifies the y-coordinate of the bullet center.
        
        text_x: int, float or callable
            Specifies the x-coordinate of the text anchor.
        
        text_y: int, float or callable
            Specifies the y-coordinate of the text anchor.
        
        text: str, callable, None or UNDEF
            Specifies the text to be drawn.
        
        text properties:
            Includes pero.TextProperties to specify the text properties.
    """
    
    bull_x = NumProperty(0)
    bull_y = NumProperty(0)
    
    text_x = NumProperty(0)
    text_y = NumProperty(0)
    
    text = StringProperty(UNDEF)
    font = Include(TextProperties)
    
    
    def get_bull_size(self, canvas, source=UNDEF, **overrides):
        """
        Gets bullet glyph width and height.
        
        Args:
            canvas: pero.Canvas
                Canvas to be used for rendering.
            
            source: any
                Data source to be used for calculating callable properties.
            
            overrides: str:any pairs
                Specific properties to be overwritten.
        
        Returns:
            (float, float)
                Object width and height.
        """
        
        raise NotImplementedError("The 'get_bull_size' method is not implemented for '%s'." % self.__class__.__name__)


class MarkerLegend(Legend):
    """
    Defines a simple legend item with marker bullet and line.
    
    Properties:
        
        show_marker: bool or callable
            Specifies whether the bullet marker should be displayed.
        
        show_line: bool or callable
            Specifies whether the bullet line should be displayed.
        
        marker: pero.MARKER, pero.Path, callable, None or UNDEF
            Specifies the marker to be used as bullet. The value can be
            specified by any item from the pero.MARKER enum or as pero.Path.
        
        marker_size: int, float or callable
            Specifies the bullet marker size.
        
        marker_line properties:
            Includes pero.LineProperties to specify the bullet marker line.
        
        marker_fill properties:
            Includes pero.FillProperties to specify the bullet marker fill.
        
        line_size: int, float or callable
            Specifies the bullet line length.
        
        line properties:
            Includes pero.LineProperties to specify the bullet line.
    """
    
    show_marker = BoolProperty(True)
    show_line = BoolProperty(False)
    
    marker = Property(UNDEF, types=(str, Path), nullable=True)
    marker_size = NumProperty(8)
    marker_line = Include(LineProperties, prefix='marker_', line_color=UNDEF)
    marker_fill = Include(FillProperties, prefix='marker_', fill_color=UNDEF)
    
    line_size = NumProperty(25)
    line = Include(LineProperties, line_color=UNDEF, line_width=UNDEF, line_style=UNDEF, line_dash=UNDEF)
    
    
    def get_bull_size(self, canvas, source=UNDEF, **overrides):
        """Gets bullet glyph size."""
        
        # check if visible
        if not self.is_visible(source, overrides):
            return None
        
        # get properties
        marker = self.get_property('marker', source, overrides)
        marker_size = self.get_property('marker_size', source, overrides)
        show_marker = self.get_property('show_marker', source, overrides)
        show_line = self.get_property('show_line', source, overrides)
        line_size = self.get_property('line_size', source, overrides)
        line_width = self.get_property('line_width', source, overrides)
        
        # init size
        width = 0
        height = 0
        
        # add marker
        if show_marker and marker_size and marker:
            if isinstance(marker, Path):
                bbox = marker.bbox()
                width = max(width, marker_size*bbox.width)
                height = max(height, marker_size*bbox.height)
            else:
                width = max(width, marker_size)
                height = max(height, marker_size)
        
        # add line
        if show_line and line_size:
            width = max(width, line_size)
            height = max(height, line_width or 0)
        
        return width, height
    
    
    def draw(self, canvas, source=UNDEF, **overrides):
        """Uses given canvas to draw legend."""
        
        # check if visible
        if not self.is_visible(source, overrides):
            return
        
        # get properties
        tag = self.get_property('tag', source, overrides)
        text = self.get_property('text', source, overrides)
        text_x = self.get_property('text_x', source, overrides)
        text_y = self.get_property('text_y', source, overrides)
        bull_x = self.get_property('bull_x', source, overrides)
        bull_y = self.get_property('bull_y', source, overrides)
        show_marker = self.get_property('show_marker', source, overrides)
        show_line = self.get_property('show_line', source, overrides)
        marker = self.get_property('marker', source, overrides)
        marker_size = self.get_property('marker_size', source, overrides)
        line_size = self.get_property('line_size', source, overrides)
        
        # start drawing group
        canvas.group(tag, "legend")
        
        # draw line
        if show_line and line_size:
            canvas.set_pen_by(self, source=source, overrides=overrides)
            x1 = bull_x - 0.5*line_size
            x2 = bull_x + 0.5*line_size
            canvas.draw_line(x1=x1, y1=bull_y, x2=x2, y2=bull_y)
        
        # draw marker
        if show_marker and marker_size and marker:
            marker_glyph = Marker.create(marker)
            marker_glyph.set_properties_from(self, src_prefix='marker_', source=source, overrides=overrides)
            marker_glyph.draw(canvas, x=bull_x, y=bull_y)
        
        # draw text
        canvas.set_text_by(self, source=source, overrides=overrides)
        canvas.draw_text(text, text_x, text_y)
        
        # end drawing group
        canvas.ungroup()


class LegendBox(Glyph):
    """
    Legend box glyph provides an easy way of drawing multiple legend items at
    given position and orientation. This can be used in case of drawing plots,
    pie charts etc.
    
    Properties:
        
        items: (pero.Legend,), callable, None or UNDEF
            Specifies a collection of legend items to draw.
        
        x: int, float or callable
            Specifies the x-coordinate of the anchor.
        
        y: int, float or callable
            Specifies the y-coordinate of the anchor.
        
        anchor: pero.POSITION_COMPASS or callable
            Specifies the anchor position within the box as any item from the
            pero.POSITION_COMPASS enum.
        
        orientation: pero.ORIENTATION or callable
            Specifies the orientation of legend items as any item from the
            pero.ORIENTATION enum.
        
        radius: int, float, (int,), (float,) callable or UNDEF
            Specifies the background box corner radius as a single value or
            values for individual corners starting from top-left.
        
        padding: int, float, (int,), (float,) callable or UNDEF
            Specifies the inner space of the background box as a single value
            or values for individual sides starting from top.
        
        spacing: int, float or callable
            Specifies the space between individual legend items.
        
        line properties:
            Includes pero.LineProperties to specify the legend background box
            outline.
        
        fill properties:
            Includes pero.FillProperties to specify the legend background box
            fill.
    """
    
    items = TupleProperty(UNDEF, types=(Legend,), nullable=True)
    
    x = NumProperty(0)
    y = NumProperty(0)
    anchor = EnumProperty(POS_NW, enum=POSITION_COMPASS)
    orientation = EnumProperty(ORI_VERTICAL, enum=ORIENTATION)
    
    radius = QuadProperty(3)
    padding = QuadProperty(5)
    spacing = NumProperty(5)
    
    line = Include(LineProperties, line_color="#ddd")
    fill = Include(FillProperties, fill_color="#fffc")
    
    
    def get_bbox(self, canvas, source=UNDEF, **overrides):
        """
        Gets glyph bounding box.
        
        Args:
            canvas: pero.Canvas
                Canvas to be used for rendering.
            
            source: any
                Data source to be used for calculating callable properties.
            
            overrides: str:any pairs
                Specific properties to be overwritten.
        
        Returns:
            pero.Frame or None
                Object bounding box.
        """
        
        # check if visible
        if not self.is_visible(source, overrides):
            return None
        
        # get items
        items = self._get_items(canvas, source, overrides)
        if not items:
            return None
        
        # get bbox
        return self._get_bbox(source, overrides, items)
    
    
    def draw(self, canvas, source=UNDEF, **overrides):
        """Uses given canvas to draw legend box."""
        
        # check if visible
        if not self.is_visible(source, overrides):
            return
        
        # get properties
        tag = self.get_property('tag', source, overrides)
        padding = self.get_property('padding', source, overrides)
        radius = self.get_property('radius', source, overrides)
        
        # get items
        items = self._get_items(canvas, source, overrides)
        if not items:
            return
        
        # get bbox
        bbox = self._get_bbox(source, overrides, items)
        x = bbox.x + padding[3]
        y = bbox.y + padding[0]
        
        # start drawing group
        canvas.group(tag, "legend_box")
        
        # draw background
        canvas.set_pen_by(self, source=source, overrides=overrides)
        canvas.set_brush_by(self, source=source, overrides=overrides)
        canvas.draw_rect(bbox.x, bbox.y, bbox.width, bbox.height, radius)
        
        # draw items
        for item, bull_bbox, text_bbox in items:
            item.draw(canvas,
                bull_x = x + bull_bbox[0] + 0.5*bull_bbox[2],
                bull_y = y + bull_bbox[1] + 0.5*bull_bbox[3],
                text_x = x + text_bbox[0],
                text_y = y + text_bbox[1])
        
        # end drawing group
        canvas.ungroup()
    
    
    def _get_items(self, canvas, source, overrides):
        """Gets initial boxes for all items."""
        
        # get properties
        items = self.get_property('items', source, overrides)
        orientation = self.get_property('orientation', source, overrides)
        spacing = self.get_property('spacing', source, overrides)
        
        # check spacing
        spacing = spacing or 0
        
        # init boxes
        boxes = []
        bull_max = 0
        for item in items:
            
            # check item
            if not item.visible or not item.text:
                continue
            
            # get sizes
            bull_w, bull_h = item.get_bull_size(canvas)
            canvas.set_text_by(item, source=source, overrides=overrides)
            text_w, text_h = canvas.get_text_size(item.text)
            line_w, line_h = canvas.get_line_size(item.text[0])
            
            # init bbox
            bull_bbox = [0, 0, bull_w, bull_h]
            text_bbox = [bull_w + spacing, 0, text_w, text_h]
            
            # center bullet with first line
            if bull_h < line_h:
                bull_bbox[1] += (line_h - bull_h)/2
            
            # add bbox
            boxes.append((item, bull_bbox, text_bbox))
            
            # keep max bullet width
            if bull_w > bull_max:
                bull_max = bull_w
        
        # align horizontally
        if orientation == ORI_HORIZONTAL:
            
            x_offset = 0
            for item, bull_bbox, text_bbox in boxes:
                bull_bbox[0] += x_offset
                text_bbox[0] += x_offset
                x_offset += bull_bbox[2] + text_bbox[2] + 4*spacing
        
        # align vertically
        else:
            
            y_offset = 0
            for item, bull_bbox, text_bbox in boxes:
                
                bull_bbox[1] += y_offset
                text_bbox[1] += y_offset
                y_offset += spacing + max(bull_bbox[3], text_bbox[3])
                
                diff = bull_max - bull_bbox[2]
                bull_bbox[0] += .5*diff
                text_bbox[0] += diff
        
        return boxes
    
    
    def _get_bbox(self, source, overrides, items):
        """Gets final bbox."""
        
        # get properties
        x = self.get_property('x', source, overrides)
        y = self.get_property('y', source, overrides)
        anchor = self.get_property('anchor', source, overrides)
        padding = self.get_property('padding', source, overrides)
        
        # init size
        width = 0
        height = 0
        
        # get width and height of all items
        for item, bull_bbox, text_bbox in items:
            width = max(width, bull_bbox[0] + bull_bbox[2], text_bbox[0] + text_bbox[2])
            height = max(height, bull_bbox[1] + bull_bbox[3], text_bbox[1] + text_bbox[3])
        
        #  apply padding
        if padding:
            width += padding[1] + padding[3]
            height += padding[0] + padding[2]
        
        # shift anchor
        if anchor == POS_NW:
            pass
        elif anchor == POS_N:
            x -= 0.5 * width
        elif anchor == POS_NE:
            x -= width
        elif anchor == POS_E:
            x -= width
            y -= 0.5 * height
        elif anchor == POS_SE:
            x -= width
            y -= height
        elif anchor == POS_S:
            x -= 0.5 * width
            y -= height
        elif anchor == POS_SW:
            y -= height
        elif anchor == POS_W:
            y -= 0.5 * height
        elif anchor == POS_C:
            x -= 0.5 * width
            y -= 0.5 * height
        
        # make bbox
        return Frame(x, y, width, height)



#  Created byMartin.cz
#  Copyright (c) Martin Strohalm. All rights reserved.

from ... enums import *


MUPDF_LINE_CAP = {
    LINE_CAP_BUTT: 0,
    LINE_CAP_ROUND: 1,
    LINE_CAP_SQUARE: 2}

MUPDF_LINE_JOIN = {
    LINE_JOIN_MITER: 0,
    LINE_JOIN_ROUND: 1,
    LINE_JOIN_BEVEL: 2}

MUPDF_LINE_STYLE = {
    LINE_STYLE_CUSTOM: 'custom',
    LINE_STYLE_SOLID: 'solid',
    LINE_STYLE_DOTTED: DASH_VALUES_DOTTED,
    LINE_STYLE_DASHED: DASH_VALUES_DASHED,
    LINE_STYLE_DASHDOTTED: DASH_VALUES_DASHDOTTED}

MUPDF_FILL_RULE = {
    FILL_RULE_EVENODD: True,
    FILL_RULE_WINDING: False}



#  Created byMartin.cz
#  Copyright (c) Martin Strohalm. All rights reserved.

import wx
from ... events import *
from .. view import View
from . enums import *
from . canvas import WXCanvas


class WXView(wx.Window, View, metaclass=type('WXViewMeta', (type(wx.Window), type(View)), {})):
    """Wrapper for wxPython View."""
    
    
    def __init__(self, parent, id=-1, size=wx.DefaultSize, style=wx.WANTS_CHARS):
        """Initializes a new instance of WXView."""
        
        # init base
        wx.Window.__init__(self, parent, id, size=size, style=style)
        View.__init__(self)
        self.SetBackgroundStyle(wx.BG_STYLE_PAINT)
        
        # init buffers
        self._dc_buffer = wx.Bitmap(*self.GetClientSize())
        self._dc_overlay = wx.Overlay()
        self._dc_overlay_empty = True
        self._use_buffer = wx.Platform == '__WXMSW__'
        
        # set window events
        self.Bind(wx.EVT_PAINT, self._on_paint)
        self.Bind(wx.EVT_SIZE, self._on_size)
        
        self.Bind(wx.EVT_KEY_DOWN, self._on_key)
        self.Bind(wx.EVT_KEY_UP, self._on_key)
        
        self.Bind(wx.EVT_MOTION, self._on_mouse)
        self.Bind(wx.EVT_MOUSEWHEEL, self._on_mouse)
        self.Bind(wx.EVT_LEAVE_WINDOW, self._on_mouse)
        
        self.Bind(wx.EVT_LEFT_DOWN, self._on_mouse)
        self.Bind(wx.EVT_LEFT_UP, self._on_mouse)
        self.Bind(wx.EVT_LEFT_DCLICK, self._on_mouse)
        
        self.Bind(wx.EVT_MIDDLE_DOWN, self._on_mouse)
        self.Bind(wx.EVT_MIDDLE_UP, self._on_mouse)
        self.Bind(wx.EVT_MIDDLE_DCLICK, self._on_mouse)
        
        self.Bind(wx.EVT_RIGHT_DOWN, self._on_mouse)
        self.Bind(wx.EVT_RIGHT_UP, self._on_mouse)
        self.Bind(wx.EVT_RIGHT_DCLICK, self._on_mouse)
    
    
    def set_cursor(self, cursor):
        """
        Sets given mouse cursor.
        
        Args:
            cursor: pero.CURSOR
                Cursor type to be set as any item from the pero.CURSOR enum.
        """
        
        # get wx cursor
        wx_cursor = cursor
        
        if wx_cursor in WX_CURSORS:
            wx_cursor = WX_CURSORS[wx_cursor]
        
        if not isinstance(wx_cursor, wx.Cursor):
            wx_cursor = wx.Cursor(wx_cursor)
        
        # set cursor
        self.SetCursor(wx_cursor)
    
    
    def set_tooltip(self, text):
        """
        Sets given text as a system tooltip.
        
        Args:
            text: str
                Tooltip text to be set.
        """
        
        self.SetToolTip(text)
    
    
    def draw_control(self):
        """Draws current control graphics."""
        
        # check control
        if self.control is None:
            return
        
        # draw into buffer
        if self._use_buffer:
            
            # init DC
            dc = wx.MemoryDC()
            dc.SelectObject(self._dc_buffer)
            
            # init canvas
            canvas = self._make_canvas(dc)
            
            # draw control
            self.control.draw(canvas)
            
            # reset overlay
            if not self._dc_overlay_empty:
                self._dc_overlay.Reset()
            
            # delete DC
            dc.SelectObject(wx.NullBitmap)
            del dc
        
        # update screen
        self.Refresh(eraseBackground=False)
        self.Update()
    
    
    def draw_overlay(self, func=None, view=None, **kwargs):
        """
        Draws cursor rubber band overlay.
        
        Specified function is expected to be called with a canvas as the first
        argument followed by given overrides (i.e. func(canvas, **kwargs)).
        If the 'func' parameter is set to None current overlay is cleared.
        
        Args:
            func: callable or None
                Drawing function to be called to draw the overlay. If set to
                None, current overlay will be cleared.
            
            view: (int, int, int, int) or None
                Rectangle defined as (x, y, width, height) used to shift origin
                of drawing canvas submitted to given drawing function.
                
            kwargs: str:any pairs
                Keyword arguments, which should be provided to the given drawing
                function.
        """
        
        # check control
        if self.control is None:
            return
        
        # do not clean if empty
        if func is None and self._dc_overlay_empty:
            return
        
        # clear current tooltip
        self.SetToolTip("")
        
        # make overlay DC
        dc = wx.ClientDC(self)
        odc = wx.DCOverlay(self._dc_overlay, dc)
        odc.Clear()
        
        # draw overlay
        if func is not None:
            
            # init canvas
            canvas = self._make_canvas(dc)
            
            # draw overlay
            if view is not None:
                with canvas.view(*view, relative=True):
                    func(canvas, **kwargs)
            else:
                func(canvas, **kwargs)
            
            # set non-empty
            self._dc_overlay_empty = False
        
        # delete DC
        del odc
    
    
    def _on_paint(self, evt):
        """Repaints current graphics."""
        
        # draw buffer on screen
        if self._use_buffer:
            wx.BufferedPaintDC(self, self._dc_buffer)
        
        # draw directly
        else:
            dc = wx.PaintDC(self)
            canvas = self._make_canvas(dc)
            self.control.draw(canvas)
            self._dc_overlay.Reset()
    
    
    def _on_size(self, evt):
        """Repaints current graphics when size has changed."""
        
        # get window size
        width, height = self.GetClientSize()
        width = max(1, width)
        height = max(1, height)
        
        # init new buffer
        if self._use_buffer:
            self._dc_buffer = wx.Bitmap(width, height)
        
        # make size event
        size_evt = SizeEvt(
            
            native = evt,
            view = self,
            control = self.control,
            
            width = width,
            height = height)
        
        # fire event
        if self.control is not None:
            self.control.fire(size_evt)
    
    
    def _on_key(self, evt):
        """Handles all key events."""
        
        # get Unicode key
        key = evt.GetUnicodeKey()
        if key != wx.WXK_NONE:
            char = chr(key)
        else:
            key = evt.GetKeyCode()
            char = None
        
        # convert to known key
        if key in WX_KEYS:
            key = WX_KEYS[key]
        
        # init base event
        key_evt = KeyEvt(
            
            native = evt,
            view = self,
            control = self.control,
            
            key = key,
            char = char,
            
            alt_down = evt.AltDown(),
            cmd_down = evt.CmdDown(),
            ctrl_down = evt.ControlDown(),
            shift_down = evt.ShiftDown())
        
        # get event type
        evt_type = evt.GetEventType()
        
        # make specific event type
        if evt_type == wx.wxEVT_KEY_DOWN:
            key_evt = KeyDownEvt.from_evt(key_evt)
        
        elif evt_type == wx.wxEVT_KEY_UP:
            key_evt = KeyUpEvt.from_evt(key_evt)
        
        # fire event
        if self.control is not None:
            self.control.fire(key_evt)
    
    
    def _on_mouse(self, evt):
        """Handles all mouse events."""
        
        # get position
        x, y = evt.GetPosition()
        
        # get wheel rotation
        if evt.GetWheelAxis() == wx.MOUSE_WHEEL_HORIZONTAL:
            x_rot = evt.GetWheelRotation()
            y_rot = 0
        else:
            x_rot = 0
            y_rot = evt.GetWheelRotation()
        
        # init base event
        mouse_evt = MouseEvt(
            
            native = evt,
            view = self,
            control = self.control,
            
            x_pos = x,
            y_pos = y,
            
            x_rot = x_rot,
            y_rot = y_rot,
            
            left_down = evt.LeftIsDown(),
            middle_down = evt.MiddleIsDown(),
            right_down = evt.RightIsDown(),
            
            alt_down = evt.AltDown(),
            cmd_down = evt.CmdDown(),
            ctrl_down = evt.ControlDown(),
            shift_down = evt.ShiftDown())
        
        # get event type
        evt_type = evt.GetEventType()
        
        # make specific event type
        if evt_type == wx.wxEVT_MOTION:
            mouse_evt = MouseMotionEvt.from_evt(mouse_evt)
        
        elif evt_type == wx.wxEVT_MOUSEWHEEL:
            mouse_evt = MouseScrollEvt.from_evt(mouse_evt)
        
        elif evt_type == wx.wxEVT_LEAVE_WINDOW:
            mouse_evt = MouseLeaveEvt.from_evt(mouse_evt)
        
        elif evt_type == wx.wxEVT_LEFT_DOWN:
            mouse_evt = LeftDownEvt.from_evt(mouse_evt)
        
        elif evt_type == wx.wxEVT_LEFT_UP:
            mouse_evt = LeftUpEvt.from_evt(mouse_evt)
        
        elif evt_type == wx.wxEVT_LEFT_DCLICK:
            mouse_evt = LeftDClickEvt.from_evt(mouse_evt)
        
        elif evt_type == wx.wxEVT_MIDDLE_DOWN:
            mouse_evt = MiddleDownEvt.from_evt(mouse_evt)
        
        elif evt_type == wx.wxEVT_MIDDLE_UP:
            mouse_evt = MiddleUpEvt.from_evt(mouse_evt)
        
        elif evt_type == wx.wxEVT_MIDDLE_DCLICK:
            mouse_evt = MiddleDClickEvt.from_evt(mouse_evt)
        
        elif evt_type == wx.wxEVT_RIGHT_DOWN:
            mouse_evt = RightDownEvt.from_evt(mouse_evt)
        
        elif evt_type == wx.wxEVT_RIGHT_UP:
            mouse_evt = RightUpEvt.from_evt(mouse_evt)
        
        elif evt_type == wx.wxEVT_RIGHT_DCLICK:
            mouse_evt = RightDClickEvt.from_evt(mouse_evt)
        
        # set focus
        if self.FindFocus() is not self and \
            evt_type in (wx.wxEVT_LEFT_DOWN, wx.wxEVT_LEFT_DCLICK,
                wx.wxEVT_MIDDLE_DOWN, wx.wxEVT_MIDDLE_DCLICK,
                wx.wxEVT_RIGHT_DOWN, wx.wxEVT_RIGHT_DCLICK):
            
            self.SetFocus()
            try: wx.Yield()
            except: pass
        
        # fire event
        if self.control is not None:
            self.control.fire(mouse_evt)
    
    
    def _make_canvas(self, dc):
        """Makes canvas using given DC."""
        
        # use GCDC
        if 'wxMac' not in wx.PlatformInfo:
            dc = wx.GCDC(dc)
        
        # make canvas
        return WXCanvas(dc)



#  Created byMartin.cz
#  Copyright (c) Martin Strohalm. All rights reserved.

import wx
from ... drawing import Graphics
from .. view import Control
from . view import WXView


class WXViewer(wx.App):
    """Simple graphics viewer application."""
    
    
    def OnInit(self):
        """Init application."""
        
        # init frame
        self._frame = WXViewFrame(None, -1, 'Pero')
        
        # show frame
        self.SetTopWindow(self._frame)
        try: wx.Yield()
        except: pass
        
        return True
    
    
    def set_size(self, size):
        """
        Sets app window size.
        
        Args:
            size: (float, float)
                App window size as (width, height).
        """
        
        self._frame.SetClientSize(size)
        self._frame.Centre(wx.BOTH)
    
    
    def set_title(self, title):
        """
        Sets app window title.
        
        Args:
            title: str
                App window title.
        """
        
        self._frame.SetTitle(title or "")
    
    
    def set_content(self, content):
        """
        Sets content to draw.
        
        Args:
            content: pero.WXView, pero.Control or pero.Graphics
                A content to set.
        """
        
        self._frame.set_content(content)
    
    
    def show(self):
        """Shows app."""
        
        self.MainLoop()
    
    
    def refresh(self):
        """Redraws view."""
        
        self._frame.refresh()
    
    
    def focus(self):
        """Sets focus to view."""
        
        self._frame.focus()


class WXViewFrame(wx.Frame):
    """Main application frame."""
    
    
    def __init__(self, parent, id, title, size=(750, 500), style=wx.DEFAULT_FRAME_STYLE | wx.NO_FULL_REPAINT_ON_RESIZE):
        
        wx.Frame.__init__(self, parent, -1, title, size=size, style=style)
        self.SetBackgroundColour((255, 255, 255))
        
        # init panel
        self._panel = wx.Panel(self, -1, style=wx.WANTS_CHARS)
        self._panel.SetSizer(wx.BoxSizer(wx.VERTICAL))
        
        # init view
        self._view = None
        
        # show frame
        self.Layout()
        self.Centre(wx.BOTH)
        self.Show(True)
        self.SetMinSize((100, 100))
    
    
    def set_content(self, content):
        """
        Sets content to draw.
        
        Args:
            content: pero.WXView, pero.Control or pero.Graphics
        """
        
        # init view
        if isinstance(content, WXView):
            self._view = content
        
        elif isinstance(content, Control):
            self._view = WXView(self._panel)
            self._view.set_control(content)
        
        elif isinstance(content, Graphics):
            self._view = WXView(self._panel)
            self._view.set_control(Control(graphics=content))
        
        else:
            message = "Unknown content type! -> %s" % type(content)
            raise TypeError(message)
        
        # clean sizer
        self._panel.Sizer.Clear(True)
        
        # add to sizer
        self._panel.Sizer.Add(self._view, 1, wx.EXPAND)
        self._panel.Sizer.Layout()
        
        # set focus
        self._view.SetFocus()
    
    
    def refresh(self):
        """Redraws view."""
        
        if self._view is not None:
            self._view.refresh()



#  Created byMartin.cz
#  Copyright (c) Martin Strohalm. All rights reserved.

from .. properties import *
from . formatter import Formatter


class EmptyFormatter(Formatter):
    """This formatter tool returns an empty string for any given value."""
    
    
    def format(self, value, *args, **kwargs):
        """
        Returns an empty string for any given value.
        
        Args:
            value: any
                Value to be formatted.
        
        Returns:
            str
                Empty string.
        """
        
        return ""


class StrFormatter(Formatter):
    """
    This formatter tool uses the newer format()-style templates to format
    given values.
    
    Properties:
        
        template: str
            Specifies the format()-style template to be used for custom
            formatting.
        
        trim: bool
            Specifies whether to automatically remove leading and trailing
            whitespace.
    """
    
    template = StringProperty("{0}", dynamic=False)
    trim = BoolProperty(True, dynamic=False, nullable=True)
    
    
    def format(self, value, *args, **kwargs):
        """
        Formats given value using custom formatting template.
        
        Args:
            value: ?
                Value to be formatted.
        
        Returns:
            str
                Formatted label.
        """
        
        # format and trim
        if self.trim:
            return self.template.format(value).strip()
        
        # format only
        return self.template.format(value)


class PrintfFormatter(Formatter):
    """
    This formatter tool uses the old %-style templates to format given values.
    
    Properties:
        
        template: str
            Specifies the %-style template to be used for custom formatting.
        
        trim: bool
            Specifies whether to automatically remove leading and trailing
            whitespace.
    """
    
    template = StringProperty('%s', dynamic=False)
    trim = BoolProperty(True, dynamic=False, nullable=True)
    
    
    def format(self, value, *args, **kwargs):
        """
        Formats given value using custom formatting template.
        
        Args:
            value: any
                Value to be formatted.
        
        Returns:
            str
                Formatted label.
        """

        # format and trim
        if self.trim:
            return (self.template % value).strip()
        
        # format only
        return self.template % value



#  Created byMartin.cz
#  Copyright (c) Martin Strohalm. All rights reserved.

from .. enums import *
from .. properties import *
from .. drawing import Graphics, Frame
from .. glyphs import Tooltip, TextTooltip
from . tool import Tool


class Control(Graphics):
    """
    Base class for all interactive controls. The main idea is to provide
    a backend-independent interface for drawing the control graphics, overlays
    and tooltips as well as assignment of specific interactivity tools for
    mouse, keyboard and touches.
    
    Properties:
        
        graphics: pero.Graphics, None or UNDEF
            Main graphics to display within the control.
    """
    
    graphics = Property(None, types=Graphics, dynamic=False, nullable=True)
    
    
    def __init__(self, **overrides):
        """Initializes a new instance of Control."""
        
        # init base
        super().__init__(**overrides)
        
        # init buffers
        self._parent = None
        self._size = (0, 0)
        
        # bind events
        self.bind(EVT_SIZE, self._on_control_size)
    
    
    def set_parent(self, parent):
        """Sets link to parent view."""
        
        # check type
        from . view import View
        if not isinstance(parent, (Control, View)):
            message = "Parent must be of type 'pero.Control' or 'pero.View'! -> %s" % type(parent)
            raise TypeError(message)
        
        # set parent
        self._parent = parent
    
    
    def set_cursor(self, cursor):
        """
        Sets given mouse cursor.
        
        Args:
            cursor: pero.CURSOR
                Cursor type to be set. The value must be an item from the
                pero.CURSOR enum.
        """
        
        # check parent
        if self._parent is None:
            return
        
        # set cursor to parent
        self._parent.set_cursor(cursor)
    
    
    def refresh(self):
        """
        Redraws current control using parent view. This makes the parent view
        responsible for initialization of a canvas and calling the 'draw' method
        to finally draw the control graphics.
        """
        
        # check parent
        if self._parent is None:
            return
        
        # refresh by parent
        self._parent.draw_control()
    
    
    def draw(self, canvas, source=UNDEF, **overrides):
        """
        Uses given canvas to draw the graphics.
        
        Args:
            canvas: pero.Canvas or None
                Specific canvas to draw the graphics on.
            
            source: any
                Data source to be used for calculating callable properties of
                current graphics.
            
            overrides: str:any pairs
                Specific properties of current graphics to be overwritten.
        """
        
        if self.graphics is not None and self.graphics is not UNDEF:
            self.graphics.draw(canvas, source=source, **overrides)
    
    
    def draw_control(self):
        """Relay control drawing to parent."""
        
        # check parent
        if self._parent is None:
            return
        
        # draw control by parent
        self._parent.draw_control()
    
    
    def draw_tooltip(self, text):
        """Relay tooltip drawing to parent."""
        
        # check parent
        if self._parent is None:
            return
        
        # draw tooltip by parent
        self._parent.draw_tooltip(text)
    
    
    def draw_overlay(self, func=None, view=None, **kwargs):
        """
        Uses parent view to initialize overlay canvas and calls given drawing
        function on it.
        
        Specified function is expected to be called with a canvas as the first
        argument followed by given arguments (i.e. func(canvas, **kwargs)).
        
        Calling this method without any parameter clears current overlay.
        
        Args:
            func: callable or None
                Drawing function to be called to draw the overlay.
            
            view: (int, int, int, int) or None
                Rectangle defined as (x, y, width, height) used to shift origin
                of drawing canvas submitted to given drawing function.
            
            kwargs: str:any pairs
                Keyword arguments, which should be provided to the given drawing
                function.
        """
        
        # check parent
        if self._parent is None:
            return
        
        # draw overlay by parent
        self._parent.draw_overlay(func, view, **kwargs)
    
    
    def clear_overlay(self):
        """Clears current overlay."""
        
        self.draw_overlay()
    
    
    def _on_control_size(self, evt):
        """Redraws current graphics when size has changed."""
        
        # get size
        self._size = (evt.width, evt.height)
        
        # draw control
        self.refresh()


class ToolControl(Control):
    """
    Main type of interactive control allowing to set specific interactivity
    tools for mouse, keyboard and touches.
    
    Properties:
        
        tooltip: pero.Tooltip, None or UNDEF
            Specifies the glyph to be used for mouse tooltip drawing. If not
            specified, a system tooltip is used instead.
        
        main_tool: pero.Tool, None or UNDEF
            Specifies the main keyboard and mouse interactivity tool. This tool
            is bound to all keyboard, mouse and touch events. Note that each
            newly assigned tool has higher priority than those currently
            assigned. Therefore, this tool should be assigned first.
        
        cursor_tool: pero.Tool, None or UNDEF
            Specifies the interactivity tool used to provide specific
            functionality for mouse cursor (i.e. no button is pressed). This
            tool is bound to all keyboard and mouse events except mouse buttons
            events. Note that each newly assigned tool has higher priority than
            those currently assigned.
        
        left_tool: pero.Tool, None or UNDEF
            Specifies the interactivity tool used to provide specific
            functionality for left mouse button. This tool is bound to all
            keyboard and mouse events including the left mouse button events.
            Note that each newly assigned tool has higher priority than those
            currently assigned.
        
        right_tool: pero.Tool, None or UNDEF
            Specifies the interactivity tool used to provide specific
            functionality for right mouse button. This tool is bound to all
            keyboard and mouse events including the right mouse button events.
            Note that each newly assigned tool has higher priority than those
            currently assigned.
        
        touch_tool: pero.Tool, None or UNDEF
            Specifies the interactivity tool used to provide specific
            functionality for touch events. This tool is bound to all keyboard
            events. Note that each newly assigned tool has higher priority than
            those currently assigned.
    """
    
    tooltip = Property(UNDEF, types=Tooltip, dynamic=False, nullable=True)
    main_tool = Property(UNDEF, types=Tool, dynamic=False, nullable=True)
    cursor_tool = Property(UNDEF, types=Tool, dynamic=False, nullable=True)
    left_tool = Property(UNDEF, types=Tool, dynamic=False, nullable=True)
    right_tool = Property(UNDEF, types=Tool, dynamic=False, nullable=True)
    touch_tool = Property(UNDEF, types=Tool, dynamic=False, nullable=True)
    
    
    def __init__(self, **overrides):
        """Initializes a new instance of Control."""
        
        # init base
        super().__init__(**overrides)
        
        # init tooltip
        if self.tooltip is UNDEF:
            self.tooltip = TextTooltip()
        
        # bind events
        self.bind(EVT_PROPERTY_CHANGED, self._on_tool_control_property_changed)
        
        # bind tools
        self._set_tool(self.main_tool, left=True, right=True, touch=True)
        self._set_tool(self.cursor_tool)
        self._set_tool(self.left_tool, left=True)
        self._set_tool(self.right_tool, right=True)
        self._set_tool(self.touch_tool, touch=True)
    
    
    def draw_tooltip(self, canvas=None, source=UNDEF, **overrides):
        """
        Draws a tooltip either by custom glyph specified by the 'tooltip'
        property or using a system tooltip of the parent view.
        
        If the custom glyph is used all the overrides are given directly to its
        drawing method (e.g. to display specific text use text='my tooltip').
        For the system tooltip actual text is also retrieved from given
        overrides assuming the 'text' key is used.
        
        If the canvas is not provided, parent view takes care to initialize it
        if necessary and the tooltip will be drawn as overlay. Therefore, this
        method can be easily used inside overlay call as well as alone.
        
        Args:
            canvas: pero.Canvas
                Canvas to draw the graphics on.
            
            source: any
                Data source to be used for calculating callable properties of
                current tooltip glyph.
            
            overrides: str:any pairs
                Specific properties of the tooltip to be overwritten.
        """
        
        # check parent
        if self._parent is None:
            return
        
        # skip tooltip
        if self.tooltip is None:
            return
        
        # use system tooltip
        if self.tooltip is UNDEF:
            self._parent.draw_tooltip(text=overrides.get('text', ""))
            return
        
        # initialize canvas
        if canvas is None:
            self.draw_overlay(self.draw_tooltip, source=source, **overrides)
            return
        
        # get clip frame
        if self.tooltip.clip is UNDEF and 'clip' not in overrides:
            overrides['clip'] = Frame(0, 0, *self._size)
        
        # draw tooltip
        self.tooltip.draw(canvas, source=source, **overrides)
    
    
    def _set_tool(self, new_tool, old_tool=None, left=False, right=False, touch=False):
        """Registers given tool."""
        
        # unbind old tool
        if old_tool:
            self.unbind(EVT_SIZE, old_tool.on_size)
            self.unbind(EVT_KEY_DOWN, old_tool.on_key_down)
            self.unbind(EVT_KEY_UP, old_tool.on_key_up)
            self.unbind(EVT_MOUSE_ENTER, old_tool.on_mouse_enter)
            self.unbind(EVT_MOUSE_LEAVE, old_tool.on_mouse_leave)
            self.unbind(EVT_MOUSE_MOTION, old_tool.on_mouse_motion)
            self.unbind(EVT_MOUSE_SCROLL, old_tool.on_mouse_scroll)
            self.unbind(EVT_LEFT_DOWN, old_tool.on_mouse_down)
            self.unbind(EVT_LEFT_UP, old_tool.on_mouse_up)
            self.unbind(EVT_LEFT_DCLICK, old_tool.on_mouse_dclick)
            self.unbind(EVT_RIGHT_DOWN, old_tool.on_mouse_down)
            self.unbind(EVT_RIGHT_UP, old_tool.on_mouse_up)
            self.unbind(EVT_TOUCH_START, old_tool.on_touch_start)
            self.unbind(EVT_TOUCH_END, old_tool.on_touch_end)
            self.unbind(EVT_TOUCH_MOVE, old_tool.on_touch_move)
            self.unbind(EVT_TOUCH_CANCEL, old_tool.on_touch_cancel)
            self.unbind(EVT_TOUCH_DTAP, old_tool.on_touch_dtap)
        
        # check tool
        if not new_tool:
            return
        
        # bind common events
        self.bind(EVT_SIZE, new_tool.on_size)
        
        # bind key events
        self.bind(EVT_KEY_DOWN, new_tool.on_key_down)
        self.bind(EVT_KEY_UP, new_tool.on_key_up)
        
        # bind main mouse events
        self.bind(EVT_MOUSE_ENTER, new_tool.on_mouse_enter)
        self.bind(EVT_MOUSE_LEAVE, new_tool.on_mouse_leave)
        self.bind(EVT_MOUSE_MOTION, new_tool.on_mouse_motion)
        self.bind(EVT_MOUSE_SCROLL, new_tool.on_mouse_scroll)
        
        # bind left mouse events
        if left:
            self.bind(EVT_LEFT_DOWN, new_tool.on_mouse_down)
            self.bind(EVT_LEFT_UP, new_tool.on_mouse_up)
            self.bind(EVT_LEFT_DCLICK, new_tool.on_mouse_dclick)
        
        # bind right mouse events
        if right:
            self.bind(EVT_RIGHT_DOWN, new_tool.on_mouse_down)
            self.bind(EVT_RIGHT_UP, new_tool.on_mouse_up)
            self.bind(EVT_RIGHT_DCLICK, new_tool.on_mouse_dclick)
        
        # bind touch events
        if touch:
            self.bind(EVT_TOUCH_START, new_tool.on_touch_start)
            self.bind(EVT_TOUCH_END, new_tool.on_touch_end)
            self.bind(EVT_TOUCH_MOVE, new_tool.on_touch_move)
            self.bind(EVT_TOUCH_CANCEL, new_tool.on_touch_cancel)
            self.bind(EVT_TOUCH_DTAP, new_tool.on_touch_dtap)
    
    
    def _on_tool_control_property_changed(self, evt):
        """Called after any property has changed."""
        
        # main tool changed
        if evt.name == 'main_tool':
            self._set_tool(evt.new_value, evt.old_value, left=True, right=True, touch=True)
        
        # cursor tool changed
        elif evt.name == 'cursor_tool':
            self._set_tool(evt.new_value, evt.old_value)
        
        # left mouse tool changed
        elif evt.name == 'left_tool':
            self._set_tool(evt.new_value, evt.old_value, left=True)
        
        # right mouse tool changed
        elif evt.name == 'right_tool':
            self._set_tool(evt.new_value, evt.old_value, right=True)
        
        # touch tool changed
        elif evt.name == 'touch_tool':
            self._set_tool(evt.new_value, evt.old_value, touch=True)



#  Created byMartin.cz
#  Copyright (c) Martin Strohalm. All rights reserved.

import math
from .. properties import *
from .. formatters import ScalarFormatter
from . utils import *
from . ticker import Ticker


class LinTicker(Ticker):
    """
    
    This type of ticker generates nice looking ticks and labels according to
    linear scale. The major and the minor steps are calculated automatically
    according to current range, however, both can be specified if needed. The
    label 'formatter' is by default set to pero.ScalarFormatter, but can be
    changed if needed.
    
    Properties:
        
        major_step: int, float or UNDEF
            Specifies the fixed step to be used for major ticks and labels. If
            not provided the value is determined automatically base on current
            range.
        
        major_count: int
            Specifies the expected number of major ticks to be generated for
            current range. The final number of ticks should be close but not
            necessary the same.
        
        major_splits: tuple of int
            Specifies the allowed splits to generate nice major ticks.
        
        minor_step: int, float or UNDEF
            Specifies the fixed step to be used for minor ticks. If not provided
            the value is determined automatically base on major ticks range.
        
        minor_count: int
            Specifies the expected number of minor ticks to be generated between
            two major ticks. The final number of ticks should be close but not
            necessary the same.
        
        minor_splits: tuple of int
            Specifies the allowed splits to generate nice minor ticks.
    """
    
    major_step = NumProperty(UNDEF, dynamic=False)
    major_count = IntProperty(7, dynamic=False)
    major_splits = TupleProperty((5, 3, 2, 1), intypes=(int, float), dynamic=False)
    
    minor_step = NumProperty(UNDEF, dynamic=False)
    minor_count = IntProperty(4, dynamic=False)
    minor_splits = TupleProperty((5, 2, 1), intypes=(int, float), dynamic=False)
    
    
    def __init__(self, **overrides):
        """Initializes a new instance of LinTicker."""
        
        # init formatter
        if 'formatter' not in overrides:
            overrides['formatter'] = ScalarFormatter()
        
        # init base
        super().__init__(**overrides)
    
    
    def make_ticks(self):
        """
        Generates ticks according to current settings.
        
        Returns:
            (float,), (float,)
                Generated major and minor ticks.
        """
        
        # get range
        start = self.formatter.convert(self.start)
        end = self.formatter.convert(self.end)
        
        # get domain
        domain = abs(end - start)
        
        # calc major step size
        major_step = self.major_step
        if major_step is UNDEF:
            major_step = calc_step_size(domain, self.major_count, self.major_splits)
        
        # calc minor step size
        minor_step = self.minor_step
        if minor_step is UNDEF:
            minor_step = calc_step_size(major_step, self.minor_count, self.minor_splits)
        
        # make ticks
        major_ticks = make_lin_ticks(start, end, major_step)
        minor_ticks = make_lin_ticks(start, end, minor_step)
        
        # invert ticks
        major_step = self.formatter.invert(major_step)
        major_ticks = [self.formatter.invert(t) for t in major_ticks]
        minor_ticks = [self.formatter.invert(t) for t in minor_ticks]
        
        # update formatter
        self.formatter.precision = abs(major_step)
        
        return major_ticks, minor_ticks
    
    
    def beautify(self, start, end):
        """
        Calculates visually nice range to cover given range.
        
        Args:
            start: int or float
                Minimum value of the range.
            
            end: int or float
                Maximum value of the range.
        
        Returns:
            (float, float)
                New extended range.
        """
        
        # get range
        start = self.formatter.convert(start)
        end = self.formatter.convert(end)
        
        # check order
        flip = False
        if start > end:
            start, end = end, start
            flip = True
        
        # calc major step size
        step = self.major_step
        if step is UNDEF:
            step = calc_step_size(abs(end - start), self.major_count, self.major_splits)
        
        # extend range
        start = math.floor(start / step) * step
        end = math.ceil(end / step) * step
        
        # invert range
        start = self.formatter.invert(start)
        end = self.formatter.invert(end)
        
        return (start, end) if not flip else (end, start)



#  Created byMartin.cz
#  Copyright (c) Martin Strohalm. All rights reserved.

from .. enums import *
from .. properties import *
from . glyph import Glyph


class ColorBar(Glyph):
    """
    Color bar provides a simple tool to visualize a color gradient range. It is
    typically used together with an axis as part of a plot.
    
    By default the gradient direction goes the same ways as device units, i.e.
    from left to right and from top to bottom. This behavior can be changed by
    setting the 'reverse' property to True.
    
    Properties:
        
        x: int, float or callable
            Specifies the x-coordinate of the top-left corner.
        
        y: int, float or callable
            Specifies the y-coordinate of the top-left corner.
        
        length: int, float or callable
            Specifies the bar length.
        
        thickness: int, float or callable
            Specifies the bar thickness.
        
        orientation: pero.ORIENTATION or callable
            Specifies the bar orientation as any item from the
            pero.ORIENTATION enum.
        
        reverse: bool or callable
            Specifies whether the gradient is drawn considering the top/left
            side (False) or the bottom/right side (True) as its start.
        
        gradient: pero.Gradient, pero.Palette, tuple, str or callable
            Specifies the color gradient as a sequence of colors,
            pero.Palette, palette name or pero.Gradient.
        
        steps: int or callable
            Specifies the number of color steps to use to draw the gradient.
        
        line properties:
            Includes pero.LineProperties to specify the outline.
        
        fill properties:
            Includes pero.FillProperties to specify the background fill.
    """
    
    x = NumProperty(0)
    y = NumProperty(0)
    length = NumProperty(0)
    thickness = NumProperty(7)
    orientation = EnumProperty(ORI_HORIZONTAL, enum=ORIENTATION)
    reverse = BoolProperty(False)
    
    gradient = GradientProperty(UNDEF)
    steps = NumProperty(128)
    
    line = Include(LineProperties, line_color="#000")
    fill = Include(FillProperties, fill_color=UNDEF)
    
    
    def draw(self, canvas, source=UNDEF, **overrides):
        """Uses given canvas to draw the bar."""
        
        # check if visible
        if not self.is_visible(source, overrides):
            return
        
        # get properties
        tag = self.get_property('tag', source, overrides)
        x = self.get_property('x', source, overrides)
        y = self.get_property('y', source, overrides)
        length = self.get_property('length', source, overrides)
        thickness = self.get_property('thickness', source, overrides)
        orientation = self.get_property('orientation', source, overrides)
        reverse = self.get_property('reverse', source, overrides)
        gradient = self.get_property('gradient', source, overrides)
        steps = self.get_property('steps', source, overrides)
        
        # get step size
        step = float(length) / steps
        
        # get direction
        direction = 1 if reverse else 0
        
        # get gradient values
        start = gradient.stops[0]
        grange = gradient.stops[-1] - start
        
        # get coords
        if orientation == ORI_HORIZONTAL:
            width = length
            height = thickness
            step_width = step
            step_height = thickness
            step_x = step
            step_y = 0
            lap_w = 0.5
            lap_h = 0
            
        else:
            width = thickness
            height = length
            step_width = thickness
            step_height = step
            step_x = 0
            step_y = step
            lap_w = 0
            lap_h = 0.5
        
        # start drawing group
        canvas.group(tag, "colorbar")
        
        # set pen and brush for background
        canvas.line_width = 0
        canvas.set_brush_by(self, source=source, overrides=overrides)
        
        # draw background
        canvas.draw_rect(x, y, width, height)
        
        # draw gradient
        for i in range(steps):
            value = start + grange * abs(direction-float(i)/steps)
            canvas.fill_color = gradient.color_at(value)
            canvas.draw_rect(x+i*step_x, y+i*step_y, step_width+lap_w, step_height+lap_h)
        
        # set pen and brush for outline
        canvas.set_pen_by(self, source=source, overrides=overrides)
        canvas.fill_color = None
        
        # draw outline
        canvas.draw_rect(x, y, width, height)
        
        # end drawing group
        canvas.ungroup()



#  Created byMartin.cz
#  Copyright (c) Martin Strohalm. All rights reserved.

import ui
from .. control import Control
from . enums import *
from . canvas import UICanvas
from . viewer import UIViewer


def show(graphics, title=None, width=None, height=None, **options):
    """
    Shows given graphics in the viewer app.
    
    Args:
        graphics: pero.Graphics or pero.Control
            Graphics to be shown.
        
        title: str or None
            Viewer frame title.
        
        width: float or None
            Viewer width in device units.
        
        height: float or None
            Viewer height in device units.
        
        style: str
            Presentation style of the ui.View. Recognized values are:
            fullscreen, sheet, popover and panel. If any touch events are
            expected to work, the fullscreen should be used.
        
        title_bar: bool
            If set to True, title bar is shown.
    """
    
    # show as image in console
    if not isinstance(graphics, Control):
        export(graphics, width=width, height=height, **options)
        return
    
    # get style
    style = "fullscreen"
    if 'style' in options:
        style = options['style']
    
    title_bar = False
    if 'title_bar' in options:
        title_bar = options['title_bar']
    
    # init main window
    window = UIViewer()
    
    # set title
    if title is not None:
        window.set_title(title)
    
    # check size
    if not width:
        width = VIEWER_WIDTH
    if not height:
        height = VIEWER_HEIGHT
    
    # set size
    window.set_size((width, height))
    
    # set graphics
    window.set_content(graphics)
    
    # draw graphics
    window.refresh()
    
    # start app
    window.present(style,
        hide_title_bar = not title_bar)


def export(graphics, path=None, width=None, height=None, **options):
    """
    Draws given graphics as raster image into specified file or into Pythonista
    console.
    
    Args:
        graphics: pero.Graphics
            Graphics to be drawn.
        
        path: str or None
            Full path of a file to save the image into. If set to None the image
            is displayed in the Pythonista console.
        
        width: float or None
            Image width in device units.
        
        height: float or None
            Image height in device units.
        
        draw_scale: float
            Drawing scaling factor.
        
        line_scale: float
            Line scaling factor.
        
        font_scale: float
            Font scaling factor.
    """
    
    # get line scale
    line_scale = 1
    if 'line_scale' in options:
        line_scale = options['line_scale']
    
    # check size
    if not width:
        width = EXPORT_WIDTH
    if not height:
        height = EXPORT_HEIGHT
    
    # open context
    with ui.ImageContext(width*line_scale, height*line_scale) as ctx:
        
        # init canvas
        canvas = UICanvas(width=width, height=height)
        
        if 'draw_scale' in options:
            canvas.draw_scale = options['draw_scale']
        
        if 'line_scale' in options:
            canvas.line_scale = options['line_scale']
        
        if 'font_scale' in options:
            canvas.font_scale = options['font_scale']
        
        # draw graphics
        graphics.draw(canvas)
        
        # get image
        img = ctx.get_image()
        
        # save to file
        if path:
            with open(path, 'wb') as f:
                f.write(img.to_png())
        
        # show image
        else:
            img.show()



#  Created byMartin.cz
#  Copyright (c) Martin Strohalm. All rights reserved.

import unittest
import pero
import numpy


class TestCase(unittest.TestCase):
    """Test case for transformation Matrix class."""
    
    
    def test_translate(self):
        """Tests whether points are translated correctly."""
        
        # translate positive
        matrix = pero.Matrix()
        matrix.translate(100, 200)
        x, y = matrix.transform(10, 20)
        
        self.assertEqual(x, 110)
        self.assertEqual(y, 220)
        
        # translate negative
        matrix = pero.Matrix()
        matrix.translate(-100, -200)
        x, y = matrix.transform(10, 20)
        
        self.assertEqual(x, -90)
        self.assertEqual(y, -180)
    
    
    def test_rotate(self):
        """Tests whether points are rotated correctly."""
        
        # rotate from zero
        matrix = pero.Matrix()
        matrix.rotate(numpy.pi*0.5)
        x, y = matrix.transform(10, 20)
        
        self.assertAlmostEqual(x, -20)
        self.assertAlmostEqual(y, 10)
        
        # rotate from origin
        matrix = pero.Matrix()
        matrix.rotate(numpy.pi*0.5, x=20, y=10)
        x, y = matrix.transform(10, 20)
        
        self.assertAlmostEqual(x, 10)
        self.assertAlmostEqual(y, 0)
        
        # rotate by degrees
        matrix = pero.Matrix()
        matrix.rotate(pero.rads(90))
        x, y = matrix.transform(10, 20)
        
        self.assertAlmostEqual(x, -20)
        self.assertAlmostEqual(y, 10)
        
        # rotate anticlockwise
        matrix = pero.Matrix()
        matrix.rotate(numpy.pi*0.5, clockwise=False)
        x, y = matrix.transform(10, 20)
        
        self.assertAlmostEqual(x, 20)
        self.assertAlmostEqual(y, -10)
    
    
    def test_scale(self):
        """Tests whether points are scaled correctly."""
        
        # scale up
        matrix = pero.Matrix()
        matrix.scale(2.5, 3)
        x, y = matrix.transform(10, 20)
        
        self.assertEqual(x, 25)
        self.assertEqual(y, 60)
        
        # scale down
        matrix = pero.Matrix()
        matrix.scale(.5, .2)
        x, y = matrix.transform(10, 20)
        
        self.assertEqual(x, 5)
        self.assertEqual(y, 4)
        
        # scale from origin
        matrix = pero.Matrix()
        matrix.scale(2, 3, x=20, y=10)
        x, y = matrix.transform(10, 20)
        
        self.assertEqual(x, 0)
        self.assertEqual(y, 40)
    
    
    def test_skew(self):
        """Tests whether points are skewed correctly."""
        
        # skew from zero
        matrix = pero.Matrix()
        matrix.skew(numpy.pi*0.25, numpy.pi*0.25)
        x, y = matrix.transform(10, 20)
        
        self.assertAlmostEqual(x, 30)
        self.assertAlmostEqual(y, 30)
        
        # skew from origin
        matrix = pero.Matrix()
        matrix.skew(numpy.pi*0.25, numpy.pi*0.25, x=20, y=10)
        x, y = matrix.transform(10, 20)
        
        self.assertAlmostEqual(x, 20)
        self.assertAlmostEqual(y, 10)
    
    
    def test_ray(self):
        """Tests whether points are rayed correctly."""
        
        # ray positive
        matrix = pero.Matrix()
        matrix.ray(10, numpy.pi/2)
        x, y = matrix.transform(10, 20)
        
        self.assertEqual(x, 10)
        self.assertEqual(y, 30)
        
        # ray negative
        matrix = pero.Matrix()
        matrix.ray(-10, numpy.pi/2)
        x, y = matrix.transform(10, 20)
        
        self.assertEqual(x, 10)
        self.assertEqual(y, 10)
    
    
    def test_hflip(self):
        """Tests whether points are flipped horizontally correctly."""
        
        # flip from zero
        matrix = pero.Matrix()
        matrix.hflip()
        x, y = matrix.transform(10, 20)
        
        self.assertEqual(x, -10)
        self.assertEqual(y, 20)
        
        # flip from origin
        matrix = pero.Matrix()
        matrix.hflip(5)
        x, y = matrix.transform(10, 20)
        
        self.assertEqual(x, 0)
        self.assertEqual(y, 20)
    
    
    def test_vflip(self):
        """Tests whether points are flipped vertically correctly."""
        
        # flip from zero
        matrix = pero.Matrix()
        matrix.vflip()
        x, y = matrix.transform(10, 20)
        
        self.assertEqual(x, 10)
        self.assertEqual(y, -20)
        
        # flip from origin
        matrix = pero.Matrix()
        matrix.vflip(10)
        x, y = matrix.transform(10, 20)
        
        self.assertEqual(x, 10)
        self.assertEqual(y, 0)


# run test case
if __name__ == "__main__":
    unittest.main(verbosity=2)



#  Created byMartin.cz
#  Copyright (c) Martin Strohalm. All rights reserved.

import os.path
import cairo
import PIL
import numpy
from . enums import *
from . canvas import CairoCanvas


def export(graphics, path, width=None, height=None, **options):
    """
    Draws given graphics into specified image file. The image format is
    determined from the extension of given file path. Supported extensions are
    .bmp, .jpg, .jpeg, .png, .pcx, .pnm, .tif, .tiff, .xpm, .ico, .cur, .svg,
    .pdf and .eps.
    
    Args:
        graphics: pero.Graphics
            Graphics to be drawn.
        
        path: str
            Full path of a file to save the image into.
        
        width: float or None
            Image width in device units.
        
        height: float or None
            Image height in device units.
        
        draw_scale: float
            Drawing scaling factor.
        
        line_scale: float
            Line scaling factor.
        
        font_scale: float
            Font scaling factor.
        
        dpi: int
            Image resolution as dots-per-inch.
        
        quality: int
            Image quality in range between 0 and 100 with 0 meaning very poor
            and 100 excellent. This option is only available for JPEG format.
    """
    
    # get filename and extension
    dirname, filename = os.path.split(path)
    basename, extension = os.path.splitext(filename)
    extension = extension.lower()
    
    # export as raster image
    if extension in CAIRO_RASTER_TYPES:
        export_raster(graphics, path, width, height, **options)
    
    # export as vector format
    elif extension in CAIRO_VECTOR_TYPES:
        export_vector(graphics, path, width, height, **options)
    
    # unsupported image format
    else:
        message = "Unsupported image format! -> %s" % extension
        raise NotImplementedError(message)


def export_raster(graphics, path, width=None, height=None, **options):
    """
    Draws given graphics as raster image into specified file. The image format
    is determined from the extension of given file path. Supported extensions
    are .bmp, .jpg, .jpeg, .png, .pcx, .tif and .tiff.
    
    Args:
        graphics: pero.Graphics
            Graphics to be drawn.
        
        path: str
            Full path of a file to save the image into.
        
        width: float or None
            Image width in device units.
        
        height: float or None
            Image height in device units.
        
        draw_scale: float
            Drawing scaling factor.
        
        line_scale: float
            Line scaling factor.
        
        font_scale: float
            Font scaling factor.
        
        dpi: int
            Image resolution as dots-per-inch.
        
        quality: int
            Image quality in range between 0 and 100 with 0 meaning very poor
            and 100 excellent. This option is only available for JPEG format.
    """
    
    # get filename and extension
    dirname, filename = os.path.split(path)
    basename, extension = os.path.splitext(filename)
    extension = extension.lower()
    
    # check format
    if extension not in CAIRO_RASTER_TYPES:
        message = "Unsupported image format! -> %s" % extension
        raise NotImplementedError(message)
    
    # check size
    if not width:
        width = EXPORT_WIDTH
    if not height:
        height = EXPORT_HEIGHT
    
    # create DC
    surface = cairo.ImageSurface(cairo.FORMAT_ARGB32, width, height)
    dc = cairo.Context(surface)
    dc.set_fill_rule(cairo.FILL_RULE_EVEN_ODD)
    
    # init canvas
    canvas = CairoCanvas(dc, width=width, height=height)
    
    if 'draw_scale' in options:
        canvas.draw_scale = options['draw_scale']
    
    if 'line_scale' in options:
        canvas.line_scale = options['line_scale']
    
    if 'font_scale' in options:
        canvas.font_scale = options['font_scale']
    
    # draw graphics
    graphics.draw(canvas)
    
    # convert to PIL image (and convert ARGB to RGBA)
    width, height = surface.get_width(), surface.get_height()
    data = numpy.frombuffer(surface.get_data(), numpy.uint8)
    data.shape = (width, height, 4)
    tmp = numpy.copy(data[:, :, 0])
    data[:, :, 0] = data[:, :, 2]
    data[:, :, 2] = tmp
    image = PIL.Image.frombuffer("RGBA", (width, height), data, 'raw', "RGBA", 0, 1)
    
    # get image options
    params = {
        'dpi': (options.get('dpi', 72), options.get('dpi', 72)),
        'quality': options.get('quality', 95)}
    
    # save to file
    image.save(path, CAIRO_RASTER_TYPES[extension], **params)


def export_vector(graphics, path, width=None, height=None, **options):
    """
    Draws given graphics as vector image into specified file. The image format
    is determined from the extension of given file path. Supported extensions
    are .svg, .pdf and .eps.
    
    Args:
        graphics: pero.Graphics
            Graphics to be drawn.
        
        path: str
            Full path of a file to save the image into.
        
        width: float or None
            Image width in device units.
        
        height: float or None
            Image height in device units.
        
        line_scale: float
            Line scaling factor.
        
        font_scale: float
            Font scaling factor.
    """
    
    # get filename and extension
    dirname, filename = os.path.split(path)
    basename, extension = os.path.splitext(filename)
    extension = extension.lower()
    
    # check format
    if extension not in CAIRO_VECTOR_TYPES:
        message = "Unsupported image format! -> %s" % extension
        raise NotImplementedError(message)
    
    # check size
    if not width:
        width = EXPORT_WIDTH
    if not height:
        height = EXPORT_HEIGHT
    
    # create surface
    if extension == '.svg':
        surface = cairo.SVGSurface(path, width, height)
    
    elif extension == '.pdf':
        surface = cairo.PDFSurface(path, width, height)
    
    elif extension == '.eps':
        surface = cairo.PSSurface(path, width, height)
        surface.set_eps(True)
    
    else:
        message = "Unsupported image format! -> %s" % extension
        raise ValueError(message)
    
    # create DC
    dc = cairo.Context(surface)
    dc.set_fill_rule(cairo.FILL_RULE_EVEN_ODD)
    
    # init canvas
    canvas = CairoCanvas(dc, width=width, height=height)
    
    if 'draw_scale' in options:
        canvas.draw_scale = options['draw_scale']
    
    if 'line_scale' in options:
        canvas.line_scale = options['line_scale']
    
    if 'font_scale' in options:
        canvas.font_scale = options['font_scale']
    
    # draw graphics
    graphics.draw(canvas)
    
    # save to file
    dc.show_page()



#  Created byMartin.cz
#  Copyright (c) Martin Strohalm. All rights reserved.

from . library import Library
from . color import Color
from . palette import Palette, PALETTES

# init library
GRADIENTS = Library()


class GradientMeta(type):
    """Defines a meta class for the main Gradient class."""
    
    
    def __getattr__(cls, key):
        """
        Gets registered named gradient by its name.
        
        Returns:
            pero.Gradient
                Corresponding gradient.
        """
        
        return GRADIENTS[key]


class Gradient(object, metaclass=GradientMeta):
    """
    Represents a gradient color generator defined by series of colors and their 
    positions.
    """
    
    
    def __init__(self, colors, stops=None, name=None):
        """
        Initializes a new instance of Gradient.
        
        Args:
            colors: (color,) or pero.Palette
                Sequence of color definitions. Any supported color definition
                can be used inside the sequence (e.g. rgba tuple, hex, name or
                pero.Color).
            
            stops: (float,) or None
                Sequence of positions for each specified color. If set to None,
                equidistant stops are generated automatically using range 0 to 1.
                If more than two colors are specified together with just two
                numbers for stop, those are used as range to automatically
                calculate equidistant stops for the range.
            
            name: str or None
                Unique name to register.
        """
        
        # set name
        self._name = name
        
        # check colors
        if len(colors) == 0:
            raise AttributeError("Gradient must be defined by one color at least!")
        elif len(colors) == 1:
            colors = (colors[0], colors[0])
        
        # get colors
        self._colors = []
        
        for color in colors:
            if not isinstance(color, Color):
                color = Color.create(color)
            self._colors.append(color)
        
        self._colors = tuple(self._colors)
        
        # check stops
        if stops is not None:
            
            # check stops type
            for stop in stops:
                if not isinstance(stop, (int, float)):
                    message = "Stop positions must be numbers! -> %s" % type(stop)
                    raise TypeError(message)
            
            # check stops sorted
            if not all(stops[i] < stops[i+1] for i in range(len(stops)-1)):
                message = "Stop positions must be sorted! -> %s" % (stops,)
                raise ValueError(message)
            
            # use range
            if len(colors) > 2 and len(stops) == 2:
                stops = Gradient.calc_stops(len(colors), stops[0], stops[1])
            
            # test same size
            elif len(colors) != len(stops):
                message = "Number of colors and stops must be equal!"
                raise ValueError(message)
        
        # generate normalized stops
        else:
            stops = Gradient.calc_stops(len(colors), 0., 1.)
        
        # set stops
        self._stops = tuple(map(float, stops))
        
        # register gradient by name
        if name is not None:
            GRADIENTS.add(self)
    
    
    @property
    def name(self):
        """
        Gets gradient name.
        
        Returns:
            str or None
                Gradient name.
        """
        
        return self._name
    
    
    @property
    def colors(self):
        """
        Gets gradient colors.
        
        Returns:
            (pero.Color,)
                Gradient colors.
        """
        
        return self._colors
    
    
    @property
    def stops(self):
        """
        Gets color stops.
        
        Returns:
            (float,)
                Positions of each color.
        """
        
        return self._stops
    
    
    def color_at(self, position, name=None):
        """
        Creates interpolated color for given position. The new color is
        automatically registered for later use if the name is specified.
        
        Args:
            position: float
                Position of the color within defined gradient range.
            
            name: str or None
                Unique name to register.
        
        Returns:
            pero.Color
                Interpolated color.
        """
        
        # check colors
        if not self._colors:
            return None
        
        # check edges
        if position <= self._stops[0]:
            return self._colors[0]
        
        if position >= self._stops[-1]:
            return self._colors[-1]
        
        # find colors around
        idx = self._locate(self._stops, position)
        left_color = self._colors[idx-1]
        right_color = self._colors[idx]
        
        # calc relative position
        left_stop = self._stops[idx-1]
        right_stop = self._stops[idx]
        position = (position - left_stop)/(right_stop - left_stop)
        
        # interpolate color
        return Color.interpolate(left_color, right_color, position, name)
    
    
    def normalized(self, start=0., end=1., name=None):
        """
        Creates a new instance of current gradient normalized to specified range.
        The new gradient is automatically registered for later use if the name
        is specified.
        
        Args:
            start: float
                Start of the normalizing range.
            
            end: float
                End of the normalizing range.
            
            name: str or None
                Unique name to register.
        
        Returns:
            pero.Gradient
                Normalized gradient.
        """
        
        # check colors
        if not self._colors:
            return Gradient(self._colors, name=name)
        
        # already normalized
        if self._stops[0] == start and self._stops[-1] == end:
            return Gradient(self._colors, self._stops, name)
        
        # calc normalization
        stops = Gradient.calc_stops(len(self._colors), start, end)
        
        # make gradient
        return Gradient(self._colors, stops, name)


    def _locate(self, items, x):
        """Locates nearest higher color."""
        
        lo = 0
        hi = len(items)
        
        while lo < hi:
            mid = (lo + hi) // 2
            if items[mid] > x:
                hi = mid
            else:
                lo = mid + 1
        
        return lo
    
    
    @staticmethod
    def create(value, name=None):
        """
        Creates new gradient from given value. The gradient can be specified
        as a sequence of color definitions, palette, gradient or palette name,
        pero.Palette or pero.Gradient. The new gradient is automatically
        registered for later use if the name is specified.
        
        Args:
            value: str, (color,), pero.Palette or pero.Gradient
                Any supported gradient definition.
            
            name: str or None
                Unique name to register.
        
        Returns:
            pero.Gradient
                Gradient.
        """
        
        # clone given gradient instance
        if isinstance(value, Gradient):
            return Gradient(value.colors, value.stops, name=name)
        
        # convert from palette
        if isinstance(value, Palette):
            return Gradient(value, name=name)
        
        # convert from color collection
        if isinstance(value, (list, tuple)):
            return Gradient(value, name=name)
        
        # convert from name
        if isinstance(value, str):
            return Gradient.from_name(value)
        
        message = "Cannot create new gradient from given value! -> %s" % (value,)
        raise ValueError(message)
    
    
    @staticmethod
    def from_name(name):
        """
        Gets the gradient from library by registered name of gradient or palette (case in-sensitive).
        
        Args:
            name: str
                Registered gradient or palette name.
        
        Returns:
            pero.Gradient
                Gradient.
        """
        
        # try from gradients
        if name in GRADIENTS:
            return GRADIENTS[name]
        
        # try from palettes
        if name in PALETTES:
            return Gradient(PALETTES[name])
        
        # name not found
        message = "Unknown gradient or palette name specified! -> '%s'" % name
        raise ValueError(message)
    
    
    @staticmethod
    def calc_stops(size, start=0., end=1.):
        """Calculates equidistant stops for given number of colors."""
        
        if size == 0:
            return ()
        
        stops = []
        step = (end - start) / (size - 1)
        
        value = start
        while value < end:
            stops.append(value)
            value += step
        
        stops.append(end)
        
        return stops



#  Created byMartin.cz
#  Copyright (c) Martin Strohalm. All rights reserved.

import numpy
from . library import Library
from . color import Color

# init library
PALETTES = Library()


class PaletteMeta(type):
    """Defines a meta class for the main Palette class."""
    
    
    def __getattr__(cls, key):
        """
        Gets registered named palette by its name.
        
        Returns:
            palette: pero.Palette
                Corresponding palette.
        """
        
        return PALETTES[key]


class Palette(object, metaclass=PaletteMeta):
    """Represents a color palette defined by series of colors."""
    
    
    def __init__(self, colors, name=None):
        """
        Initializes a new instance of Palette.
        
        Args:
            colors: (color,)
                Sequence of color definitions. Any supported color definition
                can be used inside the sequence (e.g. rgba tuple, hex, name or
                pero.Color).
            
            name: str or None
                Unique name to register.
        """
        
        # set name
        self._name = name
        
        # get colors
        self._colors = []
        for color in colors:
            if not isinstance(color, Color):
                color = Color.create(color)
            self._colors.append(color)
        
        self._colors = tuple(self._colors)
        
        # register color by name
        if name is not None:
            PALETTES.add(self)
    
    
    def __str__(self):
        """Gets standard string representation."""
        
        name = ""
        if self._name is not None:
            name = " %s" % self._name
        
        colors = [str(c) for c in self._colors]
        
        return "%s(\n\t%s)" % (name, ",\n\t".join(colors))
    
    
    def __repr__(self):
        """Gets debug string representation."""
        
        return "%s(%s)" % (self.__class__.__name__, self.__str__())
    
    
    def __len__(self):
        """Gets number of available colors."""
        
        return len(self._colors)
    
    
    def __iter__(self):
        """Gets colors iterator."""
        
        return self._colors.__iter__()
    
    
    def __getitem__(self, value):
        """Gets color at position."""
        
        return self._colors.__getitem__(value)
    
    
    @property
    def name(self):
        """
        Gets palette name.
        
        Returns:
            str or None
                Palette name.
        """
        
        return self._name
    
    
    @property
    def colors(self):
        """
        Gets colors.
        
        Returns:
            (pero.Color,)
                Palette colors.
        """
        
        return self._colors
    
    
    def lighter(self, factor=0.2, name=None):
        """
        Creates derived palette by making current colors lighter. The factor
        specifies relative amount of white to be added, i.e. 1 results in full
        white color while 0 makes no change. The new palette is automatically
        registered for later use if the name is specified.
        
        Args:
            factor: float
                Relative amount of white to be added in range 0 to 1.
            
            name: str or None
                Unique name to register.
        
        Returns:
            pero.Palette
                Lighter palette.
        """
        
        colors = [c.lighter(factor) for c in self._colors]
        return Palette(colors, name)
    
    
    def darker(self, factor=0.2, name=None):
        """
        Creates derived palette by making current colors darker. The factor
        specifies relative amount of black to be added, i.e. 1 results in full
        black color while 0 makes no change. The new palette is automatically
        registered for later use if the name is specified.
        
        Args:
            factor: float
                Relative amount of black to be added in range 0 to 1.
            
            name: str or None
                Unique name to register.
        
        Returns:
            pero.Palette
                Darker palette.
        """
        
        colors = [c.darker(factor) for c in self._colors]
        return Palette(colors, name)
    
    
    def opaque(self, opacity=1, name=None):
        """
        Creates derived palette by setting the opacity to current colors.
        0 results in fully transparent colors while 1 means fully opaque.
        The new palette is automatically registered for later use if the name
        is specified.
        
        Args:
            opacity: float
                Opacity value in range 0 to 1.
            
            name: str or None
                Unique name to register.
        
        Returns:
            pero.Palette
                New palette with specified opacity.
        """
        
        colors = [c.opaque(opacity) for c in self._colors]
        return Palette(colors, name)
    
    
    def trans(self, transparency=1, name=None):
        """
        Creates derived palette by setting the transparency to current colors.
        0 results in fully opaque colors while 1 means fully transparent.
        The new palette is automatically registered for later use if the name
        is specified.
        
        Args:
            transparency: float
                Transparency value in range 0 to 1.
            
            name: str or None
                Unique name to register.
        
        Returns:
            pero.Palette
                New palette with specified transparency.
        """
        
        colors = [c.opaque(transparency) for c in self._colors]
        return Palette(colors, name)
    
    
    def reversed(self, name=None):
        """
        Creates derived palette by taking current colors in reversed order. The
        new palette is automatically registered for later use if the name is
        specified.
        
        Args:
            name: str
                Unique name to register.
        
        Returns:
            pero.Palette
                Reversed palette.
        """
        
        return Palette(reversed(self._colors), name)
    
    
    @staticmethod
    def create(value, name=None):
        """
        Creates new palette from given value. The palette can be specified as a
        sequence of color definitions, unique library name or existing
        pero.Palette to get its copy. The new palette is automatically
        registered for later use if the name is specified.
        
        Args:
            value: str, tuple or pero.Palette
                Any supported palette definition.
            
            name: str
                Unique name to register.
        
        Returns:
            pero.Palette
                Palette.
        """
        
        # clone given palette instance
        if isinstance(value, Palette):
            return Palette(value.colors, name=name)
        
        # convert from color collection
        if isinstance(value, (list, tuple)):
            return Palette(value, name=name)
        
        # convert from name
        if isinstance(value, str):
            return Palette.from_name(value)
            
        message = "Cannot create new palette from given value! -> %s" % (value,)
        raise ValueError(message)
    
    
    @staticmethod
    def from_name(name):
        """
        Gets the palette from library by its registered name (case in-sensitive).
        
        Args:
            name: str
                Registered palette name.
        
        Returns:
            pero.Palette
                Palette.
        """
        
        # get color
        if name in PALETTES:
            return PALETTES[name]
        
        # name not found
        message = "Unknown palette name specified! -> '%s'" % name
        raise ValueError(message)
    
    
    @staticmethod
    def from_palette(palette, count, name=None):
        """
        Creates new palette by picking requested number of colors from given
        color sequence, while keeping original color range. The new palette is
        automatically registered for later use if the name is specified.
        
        Args:
            palette: pero.Palette or (pero.Color,)
                Sequence of colors to choose from.
            
            count: int
                Number of colors to pick.
            
            name: str
                Unique name to register.
        
        Returns:
            pero.Palette
                Selected colors.
        """
        
        # check count
        if count > len(palette):
            message = "Given palette contains only %s colors, %d requested!" % (len(palette), count)
            raise ValueError(message)
        
        # select colors
        selection = [int(numpy.floor(i)) for i in numpy.linspace(0, len(palette)-1, num=count)]
        colors = [palette[i] for i in selection]
        
        return Palette(colors, name)
    
    
    @staticmethod
    def from_gradient(gradient, count, name=None):
        """
        Creates new palette by interpolating requested number of colors from
        given gradient. The new palette is automatically registered for later
        use if the name is specified.
        
        Args:
            gradient: pero.Gradient, pero.Palette or (color,)
                Gradient colors.
            
            count: int
                Number of colors to pick.
            
            name: str
                Unique name to register.
        
        Returns:
            pero.Palette
                Interpolated colors.
        """
        
        # import gradient
        from .gradient import Gradient
        
        # make gradient
        if not isinstance(gradient, Gradient):
            gradient = Gradient(gradient)
        
        # select colors
        stops = [x for x in numpy.linspace(0, 1, num=count)]
        colors = [gradient.color_at(x) for x in stops]
        
        return Palette(colors, name)



#  Created byMartin.cz
#  Copyright (c) Martin Strohalm. All rights reserved.

import numpy
from .. properties import *


class Converter(PropertySet):
    """
    Converter is a simple scale-like tool to convert from one value to another
    by custom forward and reverse functions.
    
    Properties:
        
        forward: callable
            Specifies the function to be used for forward conversion. The
            function is expected to have just one argument for the input value
            and should return converted output.
        
        reverse: callable
            Specifies the function to be used for reverse conversion. The
            function is expected to have just one argument for the input value
            and should return reversed output.
    """
    
    forward = FuncProperty(UNDEF)
    reverse = FuncProperty(UNDEF)
    
    
    def scale(self, value, *args, **kwargs):
        """
        Returns corresponding converted value for given input value.
        
        Args:
            value: any or (any,)
                Input value to be scaled.
        
        Returns:
            any
                Converted value.
        """
        
        # no function defined
        if self.forward is UNDEF:
            return None
        
        # apply array scaling
        if isinstance(value, (numpy.ndarray, list, tuple)):
            return tuple(map(self.forward, value))
        
        # convert
        return self.forward(value)
    
    
    def invert(self, value, *args, **kwargs):
        """
        Returns corresponding reversed value for given output value.
        
        Args:
            value: any or (any,)
                Output value to be inverted.
        
        Returns:
            any
                Reversed value.
        """

        # no function defined
        if self.reverse is UNDEF:
            return None
        
        # apply array scaling
        if isinstance(value, (numpy.ndarray, list, tuple)):
            return tuple(map(self.reverse, value))
        
        # convert
        return self.reverse(value)



#  Created byMartin.cz
#  Copyright (c) Martin Strohalm. All rights reserved.

from .. properties import *
from . formatter import Formatter


class IndexFormatter(Formatter):
    """
    This formatter tool returns appropriate label by its index within the
    predefined set of 'labels'. A number is expected as an input of the 'format'
    method and it is automatically rounded to the nearest integer to look-up
    predefined label. If labels are not defined or the index is outside current
    range the 'default' property value is returned.
    
    Properties:
        
        labels: (str,), None or UNDEF
            Specifies the sequence of predefined labels.
        
        default: str
            Specifies the default value to be used if no appropriate label can
            be found.
    """
    
    labels = TupleProperty(UNDEF, intypes=(str,), dynamic=False, nullable=True)
    default = StringProperty("", dynamic=False, nullable=True)
    
    
    def format(self, value, *args, **kwargs):
        """
        Returns a predefined label for given index.
        
        Args:
            value: int or float
                Index value for which to look up a label.
        
        Returns:
            str
                Corresponding label.
        """
        
        # get index
        i = int(round(value))
        
        # check labels
        if not self.labels or i < 0 or i >= len(self.labels):
            return self.default
        
        # return label
        return self.labels[i]



#  Created byMartin.cz
#  Copyright (c) Martin Strohalm. All rights reserved.

from . enum import Enum

# define keys
KEY_BACK = 8
KEY_TAB = 9
KEY_RETURN = 13

KEY_SHIFT = 16
KEY_CTRL = 17
KEY_ALT = 18

KEY_ESC = 27
KEY_SPACE = 32

KEY_PAGEUP = 33
KEY_PAGEDOWN = 34
KEY_END = 35
KEY_HOME = 36
KEY_LEFT = 37
KEY_UP = 38
KEY_RIGHT = 39
KEY_DOWN = 40

KEY_DELETE = 46

KEY_0 = 48
KEY_1 = 49
KEY_2 = 50
KEY_3 = 51
KEY_4 = 52
KEY_5 = 53
KEY_6 = 54
KEY_7 = 55
KEY_8 = 56
KEY_9 = 57

KEY_A = 65
KEY_B = 66
KEY_C = 67
KEY_D = 68
KEY_E = 69
KEY_F = 70
KEY_G = 71
KEY_H = 72
KEY_I = 73
KEY_J = 74
KEY_K = 75
KEY_L = 76
KEY_M = 77
KEY_N = 78
KEY_O = 79
KEY_P = 80
KEY_Q = 81
KEY_R = 82
KEY_S = 83
KEY_T = 84
KEY_U = 85
KEY_V = 86
KEY_W = 87
KEY_X = 88
KEY_Y = 89
KEY_Z = 90

KEY_F1 = 112
KEY_F2 = 113
KEY_F3 = 114
KEY_F4 = 115
KEY_F5 = 116
KEY_F6 = 117
KEY_F7 = 118
KEY_F8 = 119
KEY_F9 = 120
KEY_F10 = 121
KEY_F11 = 122
KEY_F12 = 123

KEY = Enum(
    
    BACK = KEY_BACK,
    TAB = KEY_TAB,
    RETURN = KEY_RETURN,
    
    SHIFT = KEY_SHIFT,
    CTRL = KEY_CTRL,
    ALT = KEY_ALT,
    
    ESC = KEY_ESC,
    SPACE = KEY_SPACE,
    
    PAGEUP = KEY_PAGEUP,
    PAGEDOWN = KEY_PAGEDOWN,
    END = KEY_END,
    HOME = KEY_HOME,
    LEFT = KEY_LEFT,
    UP = KEY_UP,
    RIGHT = KEY_RIGHT,
    DOWN = KEY_DOWN,
    
    DELETE = KEY_DELETE,
    
    KEY_0 = KEY_0,
    KEY_1 = KEY_1,
    KEY_2 = KEY_2,
    KEY_3 = KEY_3,
    KEY_4 = KEY_4,
    KEY_5 = KEY_5,
    KEY_6 = KEY_6,
    KEY_7 = KEY_7,
    KEY_8 = KEY_8,
    KEY_9 = KEY_9,
    
    KEY_A = KEY_A,
    KEY_B = KEY_B,
    KEY_C = KEY_C,
    KEY_D = KEY_D,
    KEY_E = KEY_E,
    KEY_F = KEY_F,
    KEY_G = KEY_G,
    KEY_H = KEY_H,
    KEY_I = KEY_I,
    KEY_J = KEY_J,
    KEY_K = KEY_K,
    KEY_L = KEY_L,
    KEY_M = KEY_M,
    KEY_N = KEY_N,
    KEY_O = KEY_O,
    KEY_P = KEY_P,
    KEY_Q = KEY_Q,
    KEY_R = KEY_R,
    KEY_S = KEY_S,
    KEY_T = KEY_T,
    KEY_U = KEY_U,
    KEY_V = KEY_V,
    KEY_W = KEY_W,
    KEY_X = KEY_X,
    KEY_Y = KEY_Y,
    KEY_Z = KEY_Z,
    
    F1 = KEY_F1,
    F2 = KEY_F2,
    F3 = KEY_F3,
    F4 = KEY_F4,
    F5 = KEY_F5,
    F6 = KEY_F6,
    F7 = KEY_F7,
    F8 = KEY_F8,
    F9 = KEY_F9,
    F10 = KEY_F10,
    F11 = KEY_F11,
    F12 = KEY_F12)

MODIFIER_KEYS = {
    KEY_CTRL,
    KEY_SHIFT,
    KEY_ALT}



#  Created byMartin.cz
#  Copyright (c) Martin Strohalm. All rights reserved.


class Enum(object):
    """
    Defines a generic enum type, where values are provided as key:value pairs.
    The key can be used to access the value like from dict (e.g. enum[key]) or
    as property (e.g. enum.key). Each value must be unique.
    
    Note that unlike the normal dict, the Enum __contains__ method is not
    checking for keys but for values.
    """
    
    
    def __init__(self, **kwargs):
        """Initializes a new instance of Enum."""
        
        # get named values
        self._map = kwargs
        
        # get all values
        values = list(kwargs.values())
        self._values = set(values)
        
        # check values
        if len(values) != len(self._values):
            message = "Enum values are not unique! -> %s" % ", ".join(str(v) for v in values)
            raise AttributeError(message)
    
    
    def __contains__(self, value):
        """Checks whether value exists."""
        
        return value in self._values
    
    
    def __getattr__(self, name):
        """Gets value by its name."""
        
        if name in self._map:
            return self._map[name]
        
        raise AttributeError(name)
    
    
    def __getitem__(self, name):
        """Gets value by its name."""
        
        if name in self._map:
            return self._map[name]
        
        raise KeyError(name)
    
    
    def __iter__(self):
        """Gets values iterator."""
        
        return self._values.__iter__()



#  Created byMartin.cz
#  Copyright (c) Martin Strohalm. All rights reserved.

from .. properties import *


class Frame(object):
    """
    Represents a rectangular frame defined by its top-left coordinates, width
    and height.
    """
    
    
    def __init__(self, x, y, width=0, height=0):
        """
        Initializes a new instance of Frame.
        
        Args:
            x: int or float
                X-coordinate of the top left corner.
            
            y: int or float
                Y-coordinate of the top left corner.
            
            width: int or float
                Full width.
            
            height: int or float
                Full height.
        """
        
        # set values
        self._left = x
        self._top = y
        self._right = x + width
        self._bottom = y + height
        self._width = width
        self._height = height
        self._reversed = False
        
        # check values
        self._check_values()
    
    
    def __str__(self):
        """Gets standard string representation."""
        
        return "%f, %f, %f, %f" % (self._left, self._top, self._width, self._height)
    
    
    @property
    def x(self):
        """Gets x-coordinate of the left corner."""
        
        return self._left
    
    
    @property
    def y(self):
        """Gets y-coordinate of the top corner."""
        
        return self._top
    
    
    @property
    def width(self):
        """Gets full width."""
        
        return self._width
    
    
    @property
    def height(self):
        """Gets full height."""
        
        return self._height
    
    
    @property
    def left(self):
        """Gets x-coordinate of the left."""
        
        return self._left
    
    
    @property
    def right(self):
        """Gets x-coordinate of the right."""
        
        return self._right
    
    
    @property
    def top(self):
        """Gets y-coordinate of the top."""
        
        return self._top
    
    
    @property
    def bottom(self):
        """Gets y-coordinate of the bottom."""
        
        return self._bottom
    
    
    @property
    def center(self):
        """Gets coordinates of the center."""
        
        return (
            0.5 * (self._left + self._right),
            0.5 * (self._top + self._bottom))
    
    
    @property
    def x1(self):
        """Gets x-coordinate of the top left corner."""
        
        return self._left
    
    
    @property
    def y1(self):
        """Gets y-coordinate of the top left corner."""
        
        return self._top
    
    
    @property
    def x2(self):
        """Gets x-coordinate of the top right corner."""
        
        return self._right
    
    
    @property
    def y2(self):
        """Gets y-coordinate of the bottom right corner."""
        
        return self._bottom
    
    
    @property
    def cx(self):
        """Gets x-coordinate of the center."""
        
        return 0.5 * (self._left + self._right)
    
    
    @property
    def cy(self):
        """Gets y-coordinate of the center."""
        
        return 0.5 * (self._top + self._bottom)
    
    
    @property
    def tl(self):
        """Gets coordinates of the top-left corner."""
        
        return self._left, self._top
    
    
    @property
    def tr(self):
        """Gets coordinates of the top-right corner."""
        
        return self._right, self._top
    
    
    @property
    def bl(self):
        """Gets coordinates of the bottom-left corner."""
        
        return self._left, self._bottom
    
    
    @property
    def br(self):
        """Gets coordinates of the bottom-right corner."""
        
        return self._right, self._bottom
    
    
    @property
    def c(self):
        """Gets coordinates of the center."""
        
        return self.center
    
    
    @property
    def w(self):
        """Gets full width."""
        
        return self._width
    
    
    @property
    def h(self):
        """Gets full height."""
        
        return self._height
    
    
    @property
    def wh(self):
        """Gets width and height."""
        
        return self._width, self._height
    
    
    @property
    def rect(self):
        """Gets rectangle as x, y, width, height."""
        
        return self._left, self._top, self._width, self._height
    
    
    @property
    def points(self):
        """Gets rectangle as p1, p2, p3, p3 points starting from top left."""
        
        p1 = (self._left, self._top)
        p2 = (self._right, self._top)
        p3 = (self._right, self._bottom)
        p4 = (self._left, self._bottom)
        
        return p1, p2, p3, p4
    
    
    @property
    def reversed(self):
        """Returns True if the frame had originally negative width or height."""
        
        return self._reversed
    
    
    def clone(self):
        """
        Creates exact clone of current frame.
        
        Returns:
            pero.Frame
                Cloned frame.
        """
        
        frame = Frame(self._left, self._top, self._width, self._height)
        frame._reversed = self._reversed
        
        return frame
    
    
    def offset(self, x=0, y=0):
        """
        Shifts current frame by specified value in x and y directions.
        
        Args:
            x: int, float or None
                X-coordinate offset.
        
            y: int, float or None
                Y-coordinate offset.
        """
        
        # apply offset
        if x:
            self._left += x
            self._right += x
        
        if y:
            self._top += y
            self._bottom += y
        
        # check values
        self._check_values()
    
    
    def shrink(self, top=0, right=0, bottom=0, left=0):
        """
        Shrinks current frame on each specified side.
        
        Args:
            top: int, float or None
                Top padding.
            
            right: int, float or None
                Right padding.
            
            bottom: int, float or None
                Bottom padding.
            
            left: int, float or None
                Left padding.
        """
        
        # shrink values
        self._left += left
        self._right -= right
        self._top += top
        self._bottom -= bottom
        
        self._width = self._right - self._left
        self._height = self._bottom - self._top
        
        # check values
        self._check_values()
    
    
    def expand(self, top=0, right=0, bottom=0, left=0):
        """
        Expands current frame on each specified side.
        
        Args:
            top: int, float or None
                Top padding.
            
            right: int, float or None
                Right padding.
            
            bottom: int, float or None
                Bottom padding.
            
            left: int, float or None
                Left padding.
        """
        
        # expand values
        self._left -= left
        self._right += right
        self._top -= top
        self._bottom += bottom
        
        self._width = self._right - self._left
        self._height = self._bottom - self._top
        
        # check values
        self._check_values()
    
    
    def extend(self, x=None, y=None, width=0, height=0):
        """
        Extends current frame to include given coordinate, single point or 
        additional frame.
        
        Args:
            x: int, float, pero.Frame or None
                X-coordinate or frame to include.
            
            y: int, float or None
                Y-coordinate to include.
            
            width: int or float
                Width of the frame to include.
            
            height: int or float
                Height of the frame to include.
        """
        
        # get values from frame
        if isinstance(x, Frame):
            x, y, width, height = x.rect
        
        # extend width
        if x is not None:
            
            left = min(self._left, self._right, x, x+width)
            right = max(self._left, self._right, x, x+width)
            
            self._left = left
            self._right = right
            self._width = right - left
        
        # extend height
        if y is not None:
            
            top = min(self._top, self._bottom, y, y+height)
            bottom = max(self._top, self._bottom, y, y+height)
            
            self._top = top
            self._bottom = bottom
            self._height = bottom - top
        
        # check values
        self._check_values()
    
    
    def union(self, other):
        """
        Creates a new frame containing union area of current frame
        and given frame.
        
        Args:
            other: pero.Frame
                Frame to union with.
        
        Returns:
            pero.Frame or None
                Union frame.
        """
        
        # get x and width
        left = min(self._left, other._left)
        right = max(self._right, other._right)
        width = right - left
        
        # get y and height
        top = min(self._top, other._top)
        bottom = max(self._bottom, other._bottom)
        height = bottom - top
        
        return Frame(left, top, width, height)
    
    
    def intersection(self, other):
        """
        Creates a new frame containing intersection area between current frame
        and given frame or None if no such area.
        
        Args:
            other: pero.Frame
                Frame to intersect with.
        
        Returns:
            pero.Frame or None
                Intersection frame or None of no overlap.
        """
        
        # get x and width
        left = max(self._left, other._left)
        right = min(self._right, other._right)
        width = right - left
        
        if width <= 0:
            return None
        
        # get y and height
        top = max(self._top, other._top)
        bottom = min(self._bottom, other._bottom)
        height = bottom - top
        
        if height <= 0:
            return None
        
        return Frame(left, top, width, height)
    
    
    def contains(self, x, y):
        """
        Checks if given point is inside current frame.
        
        Args:
            x: int, float
                X-coordinate to check.
            
            y: int, float
                Y-coordinate to check.
        
        Returns:
            bool
                Returns True if given point is inside, False otherwise.
        """
        
        return self._left <= x <= self._right and self._top <= y <= self._bottom
    
    
    def overlaps(self, other):
        """
        Checks if there is any overlap between current frame and given frame.
        
        Args:
            other: pero.Frame
                Frame to check.
        
        Returns:
            bool
                Returns True if any overlap exists, False otherwise.
        """
        
        if not ((self._left <= other._left <= self._right)
            or (self._left <= other._right <= self._right)
            or (other._left <= self._left and other._right >= self._right)):
            return False
        
        if not ((self._top <= other._top <= self._bottom)
            or (self._top <= other._bottom <= self._bottom)
            or (other._top <= self._top and other._bottom >= self._bottom)):
            return False
        
        return True
    
    
    def _check_values(self):
        """Checks negative values."""
        
        # check width
        if self._left > self._right:
            self._left, self._right = self._right, self._left
            self._width *= -1
            self._reversed = True
        
        # check height
        if self._top > self._bottom:
            self._top, self._bottom = self._bottom, self._top
            self._height *= -1
            self._reversed = True


class FrameProperty(Property):
    """
    Defines a frame property. The value must be provided as a pero.Frame or
    as a tuple or list of four values for left x, top y, width and height.
    """
    
    
    def __init__(self, default=UNDEF, **kwargs):
        """Initializes a new instance of MarkerProperty."""
        
        kwargs['default'] = default
        kwargs['types'] = (Frame, tuple, list)
        super().__init__(**kwargs)
    
    
    def parse(self, value):
        """Validates and converts given value."""
        
        # check type
        if isinstance(value, Frame):
            return value
        
        # parse main
        value = super().parse(value)
        
        # allow UNDEF and None
        if value is UNDEF or value is None:
            return value
        
        # check func
        if callable(value):
            return value
        
        # convert to frame
        return Frame(*value)



#  Created byMartin.cz
#  Copyright (c) Martin Strohalm. All rights reserved.

import unittest
import pero


class TestCase(unittest.TestCase):
    """Test case for custom function formatter."""
    
    
    def test_formatter(self):
        """Tests whether formatter works as standalone tool."""
        
        formatter = pero.FuncFormatter(func=lambda d: "{:0.2f} u".format(d))
        
        self.assertEqual(formatter.format(0), "0.00 u")
        self.assertEqual(formatter.format(1.5), "1.50 u")
        self.assertEqual(formatter.format(1e6), "1000000.00 u")


# run test case
if __name__ == "__main__":
    unittest.main(verbosity=2)



#  Created byMartin.cz
#  Copyright (c) Martin Strohalm. All rights reserved.

# set version
version = (0, 26, 1)

# import main objects
from . enums import *
from . events import *
from . colors import Color, Palette, Gradient
from . colors import COLORS, PALETTES, GRADIENTS
from . properties import *
from . scales import *
from . formatters import *
from . tickers import *
from . drawing import *
from . glyphs import *
from . backends import *

# import modules
from . import backends
from . import colors



#  Created byMartin.cz
#  Copyright (c) Martin Strohalm. All rights reserved.

import unittest
import pero
import numpy


class TestCase(unittest.TestCase):
    """Test case for power interpolator."""
    
    
    def test_pow(self):
        """Tests whether interpolator works with 2 exponent."""
        
        interpol = pero.PowInterpol(power=2)
        
        # test inside
        self.assertAlmostEqual(interpol.normalize(50, 0, 100), 0.25, 10)
        self.assertAlmostEqual(interpol.denormalize(0.25, 0, 100), 50, 10)
        
        # test left
        self.assertAlmostEqual(interpol.normalize(0.01, 1, 100), -0.0001, 10)
        self.assertAlmostEqual(interpol.denormalize(-0.0001, 1, 100), 0.01, 10)
        
        # test right
        self.assertAlmostEqual(interpol.normalize(1000, 0, 100), 100, 10)
        self.assertAlmostEqual(interpol.denormalize(100, 0, 100), 1000, 2)
    
    
    def test_sqrt(self):
        """Tests whether interpolator works with 1/2 exponent."""
        
        interpol = pero.PowInterpol(power=0.5)
        
        # test inside
        self.assertAlmostEqual(interpol.normalize(36, 0, 100), 0.6, 10)
        self.assertAlmostEqual(interpol.denormalize(0.6, 0, 100), 36, 10)
        
        # test left
        self.assertAlmostEqual(interpol.normalize(0.01, 1, 100), -0.1, 10)
        self.assertAlmostEqual(interpol.denormalize(-0.1, 1, 100), 0.01, 10)
        
        # test right
        self.assertAlmostEqual(interpol.normalize(256, 0, 100), 1.6, 10)
        self.assertAlmostEqual(interpol.denormalize(1.6, 0, 100), 256, 10)
    
    
    def test_arrays(self):
        """Tests whether interpolator works correctly with arrays."""
        
        interpol = pero.PowInterpol(power=0.5)
        
        data = [0, 36, 256]
        model = [0, 0.6, 1.6]
        
        numpy.testing.assert_almost_equal(list(interpol.normalize(numpy.array(data), 0, 100)), model, 10)
        numpy.testing.assert_almost_equal(list(interpol.denormalize(numpy.array(model), 0, 100)), data, 10)


# run test case
if __name__ == "__main__":
    unittest.main(verbosity=2)
